"""This module contains an extended FPDF class with rudimentary HTML support"""

import fpdf_mod
import re
import os

fpdf_mod.set_global("FPDF_CACHE_MODE", 1)
fpdf_mod.set_global("SYSTEM_TTFONTS", os.path.join(os.path.dirname(__file__), "fonts"))


class FPDF_ext(fpdf_mod.FPDF):
    """This class extends FDPF to enable formatting with rudimentary HTML

    This package extends the FPDF class with the functionality to use
    rudimentary HTML for defining text formatting with the new
    method write_html()

    The class is based on the example in Tutorial 6 from
    the official FPDF documentation (http://www.fpdf.org/)
    but has extended functionality

    It's build upon the pyfpdf variant from this github:
    https://github.com/alexanderankin/pyfpdf

    Currently supports: <b>, <i>, <u>, <a>, <br>, <blockquote> <s>

    <s> is a custom tag for setting the font for part of a text. Example:
    <s fontfamily="Courier" size="14" style="B">
    Attributes can be omitted and wil then not be set.

    Unsupported tags are ignored and removed from the text
    """

    _TAB_WIDTH = 4
    _TAGS_SUPPORTED = ["b", "i", "u", "a", "br", "blockquote", "s"]

    def __init__(self, orientation="P", unit="mm", page_format="A4"):
        super().__init__(orientation=orientation, unit=unit, format=page_format)
        self._tags = dict()
        self._tags["B"] = 0
        self._tags["I"] = 0
        self._tags["U"] = 0
        self._tags["BLOCKQUOTE"] = 0
        self._href = ""
        self._last_font = None

    def write_html(self, height, html):
        """write() with support for rudimentary formatting with HTML tags"""
        html = html.replace("\n", " ")

        # split html into parts to identify all HTML tags
        # even numbered parts will contain text
        # odd numbered parts will contain tags
        parts = re.split(r"<([^>]*)>", html)

        # run through all parts one by one
        for i, part in dict(enumerate(parts)).items():
            if i % 2 == 0:
                # we have text
                if len(self._href) > 0:
                    self._put_link(self._href, height, part)
                else:
                    self.write(height, part)

            else:
                # we have a tag
                if part[0] == "/":
                    self._close_tag(part[1 : len(part)].upper())
                else:
                    # extract all attributes from the current tag if any
                    tag_parts = part.split(" ")
                    tag = tag_parts.pop(0).upper()
                    attributes = dict()
                    for tag_part in tag_parts:
                        match_obj = re.search(r'([^=]*)=["\']?([^"\']*)', tag_part)
                        if match_obj is not None and len(match_obj.groups()) == 2:
                            attributes[match_obj.group(1).upper()] = match_obj.group(2)

                    self._open_tag(tag, attributes)

    def _open_tag(self, tag, attributes):
        """set style for opening tags and singular tags"""

        if tag in ["B", "I", "U"]:
            self._set_style(tag, True)

        if tag == "BLOCKQUOTE":
            self._set_ident_plus()

        if tag == "A":
            self._href = attributes["HREF"]

        if tag == "BR":
            self.ln(5)

        if tag == "S":
            if self._last_font is not None:
                raise RuntimeError("<s> tags can not be nested")

            self._last_font = {
                "font_family": self.font_family,
                "size": self.font_size_pt,
                "style": self.font_style,
            }

            if "FONTFAMILY" in attributes:
                font_family = attributes["FONTFAMILY"]
            else:
                font_family = self.font_family

            if "SIZE" in attributes:
                size = int(attributes["SIZE"])
            else:
                size = self.font_size_pt

            if "STYLE" in attributes:
                style = attributes["STYLE"]
            else:
                style = self.font_style

            self.set_font(font_family, size=size, style=style)

    def _close_tag(self, tag):
        """set style for closing tags"""

        if tag in ["B", "I", "U"]:
            self._set_style(tag, False)

        if tag == "BLOCKQUOTE":
            self._set_ident_minus()

        if tag == "A":
            self._href = ""

        if tag == "S":
            if self._last_font is not None:
                self.set_font(
                    self._last_font["font_family"],
                    size=self._last_font["size"],
                    style=self._last_font["style"],
                )
                self._last_font = None

    def _set_style(self, tag, enable):
        """set the actual font style based on input"""
        self._tags[tag] += 1 if enable else -1
        style = ""
        for s in ["B", "I", "U"]:
            if self._tags[s] > 0:
                style += s

        self.set_font(self.font_family, size=self.font_size_pt, style=style)

    def _set_ident_plus(self):
        """moves current left margin and position forward by tab width"""
        self.set_left_margin(self.l_margin + self._TAB_WIDTH)
        self.set_x(self.get_x() + self._TAB_WIDTH)
        self.ln()

    def _set_ident_minus(self):
        """reduces current left margin and position forward by tab width"""
        left_margin = self.l_margin
        x = self.get_x()
        if left_margin > self._TAB_WIDTH and x > self._TAB_WIDTH:
            self.set_left_margin(left_margin - self._TAB_WIDTH)
            self.set_x(x - self._TAB_WIDTH)
            self.ln()

    def _put_link(self, url, height, txt):
        """ set style and write text to create a link"""
        self.set_text_color(0, 0, 255)
        self._set_style("U", True)
        self.write(height, txt, url)
        self._set_style("U", False)
        self.set_text_color(0)



# -*- coding: iso-8859-1 -*-

"PDF Template Helper for FPDF.py"

from __future__ import with_statement

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2010 Mariano Reingart"
__license__ = "LGPL 3.0"

import sys, os, csv
from .fpdf import FPDF
from .py3k import PY3K, basestring, unicode


def rgb(col):
    return (col // 65536), (col // 256 % 256), (col % 256)


class Template:
    def __init__(
        self,
        infile=None,
        elements=None,
        format="A4",
        orientation="portrait",
        title="",
        author="",
        subject="",
        creator="",
        keywords="",
    ):
        if elements:
            self.load_elements(elements)
        self.handlers = {
            "T": self.text,
            "L": self.line,
            "I": self.image,
            "B": self.rect,
            "BC": self.barcode,
            "W": self.write,
        }
        self.texts = {}
        pdf = self.pdf = FPDF(format=format, orientation=orientation, unit="mm")
        pdf.set_title(title)
        pdf.set_author(author)
        pdf.set_creator(creator)
        pdf.set_subject(subject)
        pdf.set_keywords(keywords)

    def load_elements(self, elements):
        "Initialize the internal element structures"
        self.pg_no = 0
        self.elements = elements
        self.keys = [v["name"].lower() for v in self.elements]

    def parse_csv(self, infile, delimiter=",", decimal_sep="."):
        "Parse template format csv file and create elements dict"
        keys = (
            "name",
            "type",
            "x1",
            "y1",
            "x2",
            "y2",
            "font",
            "size",
            "bold",
            "italic",
            "underline",
            "foreground",
            "background",
            "align",
            "text",
            "priority",
            "multiline",
        )
        self.elements = []
        self.pg_no = 0
        if not PY3K:
            f = open(infile, "rb")
        else:
            f = open(infile)
        with f:
            for row in csv.reader(f, delimiter=delimiter):
                kargs = {}
                for i, v in enumerate(row):
                    if not v.startswith("'") and decimal_sep != ".":
                        v = v.replace(decimal_sep, ".")
                    else:
                        v = v
                    if v == "":
                        v = None
                    else:
                        v = eval(v.strip())
                    kargs[keys[i]] = v
                self.elements.append(kargs)
        self.keys = [v["name"].lower() for v in self.elements]

    def add_page(self):
        self.pg_no += 1
        self.texts[self.pg_no] = {}

    def __setitem__(self, name, value):
        if name.lower() in self.keys:
            if not PY3K and isinstance(value, unicode):
                value = value.encode("latin1", "ignore")
            elif value is None:
                value = ""
            else:
                value = str(value)
            self.texts[self.pg_no][name.lower()] = value

    # setitem shortcut (may be further extended)
    set = __setitem__

    def has_key(self, name):
        return name.lower() in self.keys

    def __contains__(self, name):
        return self.has_key(name)

    def __getitem__(self, name):
        if name in self.keys:
            key = name.lower()
            if key in self.texts:
                # text for this page:
                return self.texts[self.pg_no][key]
            else:
                # find first element for default text:
                elements = [
                    element
                    for element in self.elements
                    if element["name"].lower() == key
                ]
                if elements:
                    return elements[0]["text"]

    def split_multicell(self, text, element_name):
        "Divide (\n) a string using a given element width"
        pdf = self.pdf
        element = [
            element
            for element in self.elements
            if element["name"].lower() == element_name.lower()
        ][0]
        style = ""
        if element["bold"]:
            style += "B"
        if element["italic"]:
            style += "I"
        if element["underline"]:
            style += "U"
        pdf.set_font(element["font"], style, element["size"])
        align = {"L": "L", "R": "R", "I": "L", "D": "R", "C": "C", "": ""}.get(
            element["align"]
        )  # D/I in spanish
        if isinstance(text, unicode) and not PY3K:
            text = text.encode("latin1", "ignore")
        else:
            text = str(text)
        return pdf.multi_cell(
            w=element["x2"] - element["x1"],
            h=element["y2"] - element["y1"],
            txt=text,
            align=align,
            split_only=True,
        )

    def render(self, outfile, dest="F"):
        pdf = self.pdf
        for pg in range(1, self.pg_no + 1):
            pdf.add_page()
            pdf.set_font("Arial", "B", 16)
            pdf.set_auto_page_break(False, margin=0)

            for element in sorted(self.elements, key=lambda x: x["priority"]):
                # print "dib",element['type'], element['name'], element['x1'], element['y1'], element['x2'], element['y2']
                element = element.copy()
                element["text"] = self.texts[pg].get(
                    element["name"].lower(), element["text"]
                )
                if "rotate" in element:
                    pdf.rotate(element["rotate"], element["x1"], element["y1"])
                self.handlers[element["type"].upper()](pdf, **element)
                if "rotate" in element:
                    pdf.rotate(0)

        if dest:
            return pdf.output(outfile, dest)

    def text(
        self,
        pdf,
        x1=0,
        y1=0,
        x2=0,
        y2=0,
        text="",
        font="arial",
        size=10,
        bold=False,
        italic=False,
        underline=False,
        align="",
        foreground=0,
        backgroud=65535,
        multiline=None,
        *args,
        **kwargs
    ):
        if text:
            if pdf.text_color != rgb(foreground):
                pdf.set_text_color(*rgb(foreground))
            if pdf.fill_color != rgb(backgroud):
                pdf.set_fill_color(*rgb(backgroud))

            font = font.strip().lower()
            if font == "arial black":
                font = "arial"
            style = ""
            for tag in "B", "I", "U":
                if text.startswith("<%s>" % tag) and text.endswith("</%s>" % tag):
                    text = text[3:-4]
                    style += tag
            if bold:
                style += "B"
            if italic:
                style += "I"
            if underline:
                style += "U"
            align = {"L": "L", "R": "R", "I": "L", "D": "R", "C": "C", "": ""}.get(
                align
            )  # D/I in spanish
            pdf.set_font(font, style, size)
            ##m_k = 72 / 2.54
            ##h = (size/m_k)
            pdf.set_xy(x1, y1)
            if multiline is None:
                # multiline==None: write without wrapping/trimming (default)
                pdf.cell(w=x2 - x1, h=y2 - y1, txt=text, border=0, ln=0, align=align)
            elif multiline:
                # multiline==True: automatic word - warp
                pdf.multi_cell(w=x2 - x1, h=y2 - y1, txt=text, border=0, align=align)
            else:
                # multiline==False: trim to fit exactly the space defined
                text = pdf.multi_cell(
                    w=x2 - x1, h=y2 - y1, txt=text, align=align, split_only=True
                )[0]
                print("trimming: *%s*" % text)
                pdf.cell(w=x2 - x1, h=y2 - y1, txt=text, border=0, ln=0, align=align)

            # pdf.Text(x=x1,y=y1,txt=text)

    def line(self, pdf, x1=0, y1=0, x2=0, y2=0, size=0, foreground=0, *args, **kwargs):
        if pdf.draw_color != rgb(foreground):
            # print "SetDrawColor", hex(foreground)
            pdf.set_draw_color(*rgb(foreground))
        # print "SetLineWidth", size
        pdf.set_line_width(size)
        pdf.line(x1, y1, x2, y2)

    def rect(
        self,
        pdf,
        x1=0,
        y1=0,
        x2=0,
        y2=0,
        size=0,
        foreground=0,
        backgroud=65535,
        *args,
        **kwargs
    ):
        if pdf.draw_color != rgb(foreground):
            pdf.set_draw_color(*rgb(foreground))
        if pdf.fill_color != rgb(backgroud):
            pdf.set_fill_color(*rgb(backgroud))
        pdf.set_line_width(size)
        pdf.rect(x1, y1, x2 - x1, y2 - y1)

    def image(self, pdf, x1=0, y1=0, x2=0, y2=0, text="", *args, **kwargs):
        if text:
            pdf.image(text, x1, y1, w=x2 - x1, h=y2 - y1, type="", link="")

    def barcode(
        self,
        pdf,
        x1=0,
        y1=0,
        x2=0,
        y2=0,
        text="",
        font="arial",
        size=1,
        foreground=0,
        *args,
        **kwargs
    ):
        if pdf.draw_color != rgb(foreground):
            pdf.set_draw_color(*rgb(foreground))
        font = font.lower().strip()
        if font == "interleaved 2of5 nt":
            pdf.interleaved2of5(text, x1, y1, w=size, h=y2 - y1)

    # Added by Derek Schwalenberg Schwalenberg1013@gmail.com to allow (url) links in templates (using write method) 2014-02-22
    def write(
        self,
        pdf,
        x1=0,
        y1=0,
        x2=0,
        y2=0,
        text="",
        font="arial",
        size=1,
        bold=False,
        italic=False,
        underline=False,
        align="",
        link="http://example.com",
        foreground=0,
        *args,
        **kwargs
    ):
        if pdf.text_color != rgb(foreground):
            pdf.set_text_color(*rgb(foreground))
        font = font.strip().lower()
        if font == "arial black":
            font = "arial"
        style = ""
        for tag in "B", "I", "U":
            if text.startswith("<%s>" % tag) and text.endswith("</%s>" % tag):
                text = text[3:-4]
                style += tag
        if bold:
            style += "B"
        if italic:
            style += "I"
        if underline:
            style += "U"
        align = {"L": "L", "R": "R", "I": "L", "D": "R", "C": "C", "": ""}.get(
            align
        )  # D/I in spanish
        pdf.set_font(font, style, size)
        ##m_k = 72 / 2.54
        ##h = (size/m_k)
        pdf.set_xy(x1, y1)
        pdf.write(5, text, link)



from .fpdf_ext import FPDF_ext
from . import settings


class MyFPDF(FPDF_ext):
    """Inheritance of FPDF class to add header and footers and set PDF settings"""

    def __init__(self, orientation="P", unit="mm", page_format="A4"):
        super().__init__(orientation=orientation, unit=unit, page_format=page_format)
        self._page_title = ""

    @property
    def page_title(self):
        """text shown as title on every page"""
        return self._page_title

    @page_title.setter
    def page_title(self, text):
        """set text to appear as title on every page"""
        self._page_title = str(text)

    def header(self):
        """definition of custom header"""
        self.set_font(
            settings.FONT_FAMILY_DEFAULT, size=settings.FONT_SIZE_NORMAL, style="B"
        )
        self.cell(0, 0, self._page_title, 0, 1, "C")
        self.ln(settings.LINE_HEIGHT_DEFAULT)

    def footer(self):
        """definition of custom footer"""
        self.set_y(-15)
        self.cell(0, 10, "Page " + str(self.page_no()) + " / {nb}", 0, 0, "C")

    def write_info_table(self, table_def):
        """write info table defined by dict"""
        cell_height = 10
        for key, value in table_def.items():
            self.set_font(self.font_family, style="B")
            self.cell(50, cell_height, str(key), 1)
            self.set_font(self.font_family)
            self.cell(0, cell_height, str(value), 1)
            self.ln()



"""Handles all requests to Slack"""

import logging
from time import sleep

from babel.numbers import format_decimal
import slack

from . import settings
from .helpers import transform_encoding
from .locales import LocaleHelper


logger = logging.getLogger(__name__)


class SlackService:
    """Service layer between main app and Slack API"""

    def __init__(self, slack_token: str, locale_helper: LocaleHelper = None) -> None:
        """
        Args:
        - slack_token: Slack token to use for all API calls
        - locale_helper: locale to use
        """
        if slack_token is None:
            raise ValueError("slack_token can not be null")

        # load information for current Slack workspace
        self._client = slack.WebClient(token=slack_token)
        if not locale_helper:
            locale_helper = LocaleHelper()
        self._locale = locale_helper.locale
        if slack_token != "TEST":
            self._workspace_info = self._fetch_workspace_info()
            logger.info("Current Slack workspace: %s", self.team)
            self._user_names = self.fetch_user_names()

            # set author
            if "user_id" in self._workspace_info:
                author_id = self._workspace_info["user_id"]
                if self._workspace_info["user_id"] in self._user_names:
                    self._author = self._user_names[author_id]
                else:
                    self._author = "unknown_user_" + self._workspace_info["user_id"]
            else:
                author_id = None
                self._author = "unknown user"

            logger.info("Current Slack user: %s", self.author)
            self._channel_names = self._fetch_channel_names()
            self._usergroup_names = self._fetch_usergroup_names()
            self._is_test_mode = False

        else:
            # if started with TEST parameter class properties will be
            # initialized empty and need to be set manually in test setup
            self._workspace_info = dict()
            self._user_names = dict()
            self._channel_names = dict()
            self._usergroup_names = dict()
            author_id = None
            self._author = "test user"
            self._is_test_mode = True

        if author_id is not None:
            self._author_info = self._fetch_user_info(author_id)
        else:
            self._author_info = dict()

    @property
    def author(self) -> str:
        return self._author

    @property
    def team(self) -> str:
        return self._workspace_info["team"]

    @property
    def is_test_mode(self) -> bool:
        return self._is_test_mode

    def author_info(self) -> dict:
        return self._author_info

    def channel_names(self) -> dict:
        return self._channel_names

    def user_names(self) -> dict:
        return self._user_names

    def usergroup_names(self) -> dict:
        return self._usergroup_names

    def _fetch_workspace_info(self) -> dict:
        """returns dict with info about current workspace"""

        logger.info("Fetching workspace info from Slack...")
        res = self._client.auth_test()
        return res.data

    def fetch_user_names(self) -> dict:
        """returns dict of user names with user ID as key"""
        user_names_raw = self._fetch_pages(
            "users_list", key="members", items_name="users"
        )
        user_names = self._reduce_to_dict(user_names_raw, "id", "real_name", "name")
        for user in user_names:
            user_names[user] = transform_encoding(user_names[user])
        return user_names

    def _fetch_user_info(self, user_id: str) -> dict:
        """returns dict of user info for user ID incl. locale"""
        logger.info("Fetching user info for author...")
        response = self._client.users_info(user=user_id, include_locale=1)
        return response["user"]

    def _fetch_channel_names(self) -> dict:
        """returns dict of channel names with channel ID as key"""
        channel_names_raw = self._fetch_pages(
            "conversations_list",
            key="channels",
            args={"types": "public_channel,private_channel"},
            items_name="channels",
        )
        channel_names = self._reduce_to_dict(channel_names_raw, "id", "name")
        for channel in channel_names:
            channel_names[channel] = transform_encoding(channel_names[channel])
        return channel_names

    def _fetch_usergroup_names(self) -> dict:
        """returns dict of usergroup names with usergroup ID as key"""

        logger.info("Fetching usergroups from Slack...")
        response = self._client.usergroups_list()
        usergroup_names = self._reduce_to_dict(response["usergroups"], "id", "handle")
        if usergroup_names:
            for usergroup in usergroup_names:
                usergroup_names[usergroup] = transform_encoding(
                    usergroup_names[usergroup]
                )
            logger.info(
                "Got a total of %s usergroups for this workspace",
                format_decimal(len(usergroup_names), locale=self._locale),
            )
        else:
            logger.info("This workspace has no usergroups")
        return usergroup_names

    def fetch_messages_from_channel(
        self, channel_id, max_messages, oldest=None, latest=None
    ) -> list:
        """retrieve messages from a channel on Slack and return as list"""

        oldest_ts = str(oldest.timestamp()) if oldest is not None else 0
        latest_ts = str(latest.timestamp()) if latest is not None else 0
        messages = self._fetch_pages(
            "conversations_history",
            key="messages",
            args={
                "channel": channel_id,
                "oldest": oldest_ts,
                "latest": latest_ts,
            },
            max_rows=max_messages,
            items_name="messages",
            collection_name="channel",
        )
        return messages

    def fetch_threads_from_messages(
        self, channel_id, messages, max_messages, oldest=None, latest=None
    ) -> dict:
        """returns threads from all messages from for a channel as dict"""
        threads = dict()
        thread_num = 0
        thread_messages_total = 0
        for msg in messages:
            if "thread_ts" in msg and msg["thread_ts"] == msg["ts"]:
                thread_ts = msg["thread_ts"]
                thread_num += 1
                thread_messages = self._fetch_messages_from_thread(
                    channel_id, thread_ts, max_messages, oldest, latest
                )
                threads[thread_ts] = thread_messages
                thread_messages_total += len(thread_messages)

        if thread_messages_total:
            logger.info(
                "Received %s messages from %d threads",
                format_decimal(thread_messages_total, locale=self._locale),
                thread_num,
            )
        else:
            logger.info("This channel has no threads")

        return threads

    def _fetch_messages_from_thread(
        self, channel_id, thread_ts, max_messages, oldest=None, latest=None
    ) -> list:
        """retrieve messages from a Slack thread and return as list"""
        oldest_ts = str(oldest.timestamp()) if oldest is not None else 0
        latest_ts = str(latest.timestamp()) if latest is not None else 0
        messages = self._fetch_pages(
            "conversations_replies",
            key="messages",
            args={
                "channel": channel_id,
                "ts": thread_ts,
                "oldest": oldest_ts,
                "latest": latest_ts,
            },
            max_rows=max_messages,
            items_name="threads",
            collection_name="channel",
            print_result=False,
        )
        return messages

    def _fetch_pages(
        self,
        method,
        key: str,
        args: dict = None,
        limit: int = None,
        max_rows: int = None,
        items_name: str = None,
        collection_name: str = None,
        print_result: bool = True,
    ) -> list:
        """helper for retrieving all pages from an API endpoint"""
        # fetch first page
        page = 1
        output_str = (
            f"Fetching {items_name if items_name else method} "
            f"from {collection_name if collection_name else 'workspace'}..."
        )
        logger.info(output_str)
        if not args:
            args = {}
        if not limit:
            limit = settings.SLACK_PAGE_LIMIT
        base_args = {**args, **{"limit": limit}}
        response = getattr(self._client, method)(**base_args)
        rows = response[key]

        # fetch additional page (if any)
        while (
            (not max_rows or len(rows) < max_rows)
            and response.get("response_metadata")
            and response["response_metadata"].get("next_cursor")
        ):
            page += 1
            logger.info("%s - page %s", output_str, page)
            sleep(1)  # need to wait 1 sec before next call due to rate limits
            page_args = {
                **base_args,
                **{
                    "cursor": response["response_metadata"].get("next_cursor"),
                },
            }
            response = getattr(self._client, method)(**page_args)
            rows += response[key]

        if print_result:
            logger.info(
                "Received %s %s",
                format_decimal(len(rows), locale=self._locale),
                items_name if items_name else "objects",
            )
        return rows

    def fetch_bot_names_for_messages(self, messages: list, threads: list) -> dict:
        """Fetches bot names from API for provided messages

        Will only fetch names for bots that never appeared with a username
        in any message (lazy approach since calls to bots_info are very slow)
        """
        # collect bot_ids without user name from messages
        bot_ids = list()
        bot_names = dict()
        for msg in messages:
            if "bot_id" in msg:
                bot_id = msg["bot_id"]
                if "username" in msg:
                    bot_names[bot_id] = transform_encoding(msg["username"])
                else:
                    bot_ids.append(bot_id)

        # collect bot_ids without user name from thread messages
        for thread_messages in threads:
            for msg in thread_messages:
                if "bot_id" in msg:
                    bot_id = msg["bot_id"]
                    if "username" in msg:
                        bot_names[bot_id] = transform_encoding(msg["username"])
                    else:
                        bot_ids.append(bot_id)

        # Find bot IDs that are not in bot_names
        bot_ids = set(bot_ids).difference(bot_names.keys())

        # collect bot names from API if needed
        if len(bot_ids) > 0:
            logger.info("Fetching names for %d bots", len(bot_ids))
            for bot_id in bot_ids:
                response = self._client.bots_info(bot=bot_id)
                if response["ok"]:
                    bot_names[bot_id] = transform_encoding(response["bot"]["name"])
                    sleep(1)  # need to wait 1 sec before next call due to rate limits

        return bot_names

    @staticmethod
    def _reduce_to_dict(
        arr: list, key_name: str, col_name_primary: str, col_name_secondary: str = None
    ) -> list:
        """returns dict with selected columns as key and value from list of dict

        Args:
            arr: list of dicts to reduce
            key_name: name of column to become key
            col_name_primary: colum will become value if it exists
            col_name_secondary: colum will become value if col_name_primary
                does not exist and this argument is provided

        dict items with no matching key_name, col_name_primary and
        col_name_secondary will not be included in the resulting new dict

        """
        arr2 = dict()
        for item in arr:
            if key_name in item:
                key = item[key_name]
                if col_name_primary in item:
                    arr2[key] = item[col_name_primary]
                elif col_name_secondary is not None and col_name_secondary in item:
                    arr2[key] = item[col_name_secondary]
        return arr2



#!/usr/bin/env python
# -*- coding: latin-1 -*-

from .py3k import PY3K, basestring, unicode

# fpdf php helpers:


def substr(s, start, length=-1):
    if length < 0:
        length = len(s) - start
    return s[start : start + length]


def sprintf(fmt, *args):
    return fmt % args


def print_r(array):
    if not isinstance(array, dict):
        array = dict([(k, k) for k in array])
    for k, v in array.items():
        print("[%s] => %s " % (k, v))


def UTF8ToUTF16BE(instr, setbom=True):
    "Converts UTF-8 strings to UTF16-BE."
    outstr = "".encode()
    if setbom:
        outstr += "\xFE\xFF".encode("latin1")
    if not isinstance(instr, unicode):
        instr = instr.decode("UTF-8")
    outstr += instr.encode("UTF-16BE")
    # convert bytes back to fake unicode string until PEP461-like is implemented
    if PY3K:
        outstr = outstr.decode("latin1")
    return outstr


def UTF8StringToArray(instr):
    "Converts UTF-8 strings to codepoints array"
    return [ord(c) for c in instr]


# ttfints php helpers:


def die(msg):
    raise RuntimeError(msg)


def str_repeat(s, count):
    return s * count


def str_pad(s, pad_length=0, pad_char=" ", pad_type=+1):
    if pad_type < 0:  # pad left
        return s.rjust(pad_length, pad_char)
    elif pad_type > 0:  # pad right
        return s.ljust(pad_length, pad_char)
    else:  # pad both
        return s.center(pad_length, pad_char)


strlen = count = lambda s: len(s)



"""command line interface"""

import argparse
import os
from pathlib import Path
import sys

from babel import Locale, UnknownLocaleError
from dateutil import parser
import pytz
from slack.errors import SlackApiError

from . import __version__
from . import settings
from .channel_exporter import SlackChannelExporter


def main():
    """Implements the arg parser and starts the channel exporter with its input"""

    args = parse_args(sys.argv[1:])
    if "version" in args:
        print(__version__)
        exit(0)

    # try to take slack token from optional argument or environment variable
    if args.token is None:
        if "SLACK_TOKEN" in os.environ:
            slack_token = os.environ["SLACK_TOKEN"]
        else:
            print("ERROR: No slack token provided")
            exit(1)
    else:
        slack_token = args.token

    # parse local timezone
    if args.timezone is not None:
        try:
            my_tz = pytz.timezone(args.timezone)
        except pytz.UnknownTimeZoneError:
            print("ERROR: Unknown timezone")
            exit(1)
    else:
        my_tz = None

    # parse locale
    if args.locale is not None:
        try:
            my_locale = Locale.parse(args.locale, sep="-")
        except UnknownLocaleError:
            print("ERROR: provided locale string is not valid")
            exit(1)
    else:
        my_locale = None

    # parse oldest
    if args.oldest is not None:
        try:
            oldest = parser.parse(args.oldest)
        except ValueError:
            print("ERROR: Invalid date input for --oldest")
            exit(1)
    else:
        oldest = None

    # parse latest
    if args.latest is not None:
        try:
            latest = parser.parse(args.latest)
        except ValueError:
            print("ERROR: Invalid date input for --latest")
            exit(1)
    else:
        latest = None

    if not args.quiet:
        channel_postfix = "s" if args.channel and len(args.channel) > 1 else ""
        print(f"Exporting channel{channel_postfix} from Slack...")
    try:
        exporter = SlackChannelExporter(
            slack_token=slack_token,
            my_tz=my_tz,
            my_locale=my_locale,
            add_debug_info=args.add_debug_info,
        )
    except SlackApiError as ex:
        print(f"ERROR: {ex}")
        exit(1)

    result = exporter.run(
        channel_inputs=args.channel,
        dest_path=Path(args.destination) if args.destination else None,
        oldest=oldest,
        latest=latest,
        page_orientation=args.page_orientation,
        page_format=args.page_format,
        max_messages=args.max_messages,
        write_raw_data=(args.write_raw_data is True),
    )
    for channel in result["channels"].values():
        if not args.quiet:
            print(
                f"{'written' if channel['ok'] else 'failed'}: {channel['filename_pdf']}"
            )


def parse_args(args: list) -> argparse.ArgumentParser:
    """defines the argument parser and returns parsed result from given argument"""
    my_arg_parser = argparse.ArgumentParser(
        description="This program exports the text of a Slack channel to a PDF file",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    # main arguments
    my_arg_parser.add_argument(
        "channel", help="One or several: name or ID of channel to export.", nargs="+"
    )
    my_arg_parser.add_argument("--token", help="Slack OAuth token")
    my_arg_parser.add_argument("--oldest", help="don't load messages older than a date")
    my_arg_parser.add_argument("--latest", help="don't load messages newer then a date")

    # PDF file
    my_arg_parser.add_argument(
        "-d",
        "--destination",
        help="Specify a destination path to store the PDF file. (TBD)",
        default=".",
    )

    # formatting
    my_arg_parser.add_argument(
        "--page-orientation",
        help="Orientation of PDF pages",
        choices=["portrait", "landscape"],
        default=settings.PAGE_ORIENTATION_DEFAULT,
    )
    my_arg_parser.add_argument(
        "--page-format",
        help="Format of PDF pages",
        choices=["a3", "a4", "a5", "letter", "legal"],
        default=settings.PAGE_FORMAT_DEFAULT,
    )
    my_arg_parser.add_argument(
        "--timezone",
        help=(
            "Manually set the timezone to be used e.g. 'Europe/Berlin' "
            "Use a timezone name as defined here: "
            "https://en.wikipedia.org/wiki/List_of_tz_database_time_zones"
        ),
    )
    my_arg_parser.add_argument(
        "--locale",
        help=(
            "Manually set the locale to be used with a IETF language tag, "
            "e.g. ' de-DE' for Germany. "
            "See this page for a list of valid tags: "
            "https://en.wikipedia.org/wiki/IETF_language_tag"
        ),
    )

    # standards
    my_arg_parser.add_argument(
        "--version",
        help="show the program version and exit",
        action="version",
        version=__version__,
    )

    # exporter config
    my_arg_parser.add_argument(
        "--max-messages",
        help="max number of messages to export",
        type=int,
        default=settings.MAX_MESSAGES_PER_CHANNEL,
    )

    # Developer needs
    my_arg_parser.add_argument(
        "--write-raw-data",
        help=(
            "will also write all raw data returned from the API to files,"
            " e.g. messages.json with all messages"
        ),
        action="store_const",
        const=True,
    )
    my_arg_parser.add_argument(
        "--add-debug-info",
        help="wether to add debug info to PDF",
        action="store_const",
        const=True,
        default=False,
    )
    my_arg_parser.add_argument(
        "--quiet",
        action="store_const",
        const=True,
        default=False,
        help=(
            "When provided will not generate normal console output, "
            "but still show errors "
            "(console logging not affected and needs to be configured through "
            "log levels instead)"
        ),
    )
    return my_arg_parser.parse_args(args)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"FPDF for python"

__license__ = "LGPL 3.0"
__version__ = "1.7.2"

from .fpdf import (
    FPDF,
    FPDF_FONT_DIR,
    FPDF_VERSION,
    SYSTEM_TTFONTS,
    set_global,
    FPDF_CACHE_MODE,
    FPDF_CACHE_DIR,
)

try:
    from .html import HTMLMixin
except ImportError:
    import warnings

    warnings.warn("web2py gluon package not installed, required for html2pdf")

from .template import Template



import re

from .helpers import transform_encoding
from .locales import LocaleHelper
from .slack_service import SlackService


class MessageTransformer:
    """parsing and transforming Slack messages"""

    def __init__(
        self,
        slack_service: SlackService,
        locale_helper: LocaleHelper,
        font_family_mono_default: str,
    ) -> None:
        self._slack_service = slack_service
        self._locale_helper = locale_helper
        self._font_family_mono_default = font_family_mono_default

    def transform_text(self, text: str, use_mrkdwn: bool = False) -> str:
        """transforms mrkdwn text into HTML text for PDF output

        Main method to resolve all mrkdwn, e.g. <C12345678>, <!here>, *bold*
        Will resolve channel and user IDs to their names if possible
        Returns string with rudimentary HTML for formatting and links

        Attr:
            text: text string to be transformed
            use_mrkdwn: will transform mrkdwn if set to true

        Returns:
            transformed text string with HTML formatting
        """

        # pass 1 - adjust encoding to latin-1 and transform HTML entities
        s2 = transform_encoding(text)

        # if requested try to transform mrkdwn in text
        if use_mrkdwn:

            # pass 2 - transform mrkdwns with brackets
            s2 = re.sub(r"<(.*?)>", self._replace_mrkdwn_in_text, s2)

            # pass 3 - transform formatting mrkdwns

            # bold
            s2 = re.sub(r"\*(.+)\*", r"<b>\1</b>", s2)
            # italic
            s2 = re.sub(r"\b_(.+)_\b", r"<i>\1</i>", s2)
            # code
            s2 = re.sub(
                r"`(.*)`",
                r'<s fontfamily="' + self._font_family_mono_default + r'">\1</s>',
                s2,
            )
            # indents
            s2 = re.sub(r"^>(.+)", r"<blockquote>\1</blockquote>", s2, 0, re.MULTILINE)
            s2 = s2.replace("</blockquote><br>", "</blockquote>")
            # EOF
            s2 = s2.replace("\n", "<br>")

        return s2

    def _replace_mrkdwn_in_text(self, matchObj: re.Match) -> str:
        """inline function returns replacement string for re.sub

        This function does the actual resolving of IDs and mrkdwn key words
        """
        match = matchObj.group(1)

        id_chars = match[0:2]
        id_raw = match[1 : len(match)]
        parts = id_raw.split("|", 1)
        obj_id = parts[0]

        make_bold = True
        if id_chars == "@U" or id_chars == "@W":
            # match is a user ID
            if obj_id in self._slack_service.user_names():
                replacement = "@" + self._slack_service.user_names()[obj_id]
            else:
                replacement = f"@user_{obj_id}"

        elif id_chars == "#C":
            # match is a channel ID
            if obj_id in self._slack_service.channel_names():
                replacement = "#" + self._slack_service.channel_names()[obj_id]
            else:
                replacement = f"#channel_{obj_id}"

        elif match[0:9] == "!subteam^":
            # match is a user group ID
            match2 = re.match(r"!subteam\^(S[A-Z0-9]+)", match)
            if match2 is not None and len(match2.groups()) == 1:
                usergroup_id = match2.group(1)
                if usergroup_id in self._slack_service.usergroup_names():
                    usergroup_name = self._slack_service.usergroup_names()[usergroup_id]
                else:
                    usergroup_name = f"usergroup_{usergroup_id}"
            else:
                usergroup_name = "usergroup_unknown"
            replacement = "@" + usergroup_name

        elif match[0:1] == "!":
            # match is a special mention
            if obj_id == "here":
                replacement = "@here"

            elif obj_id == "channel":
                replacement = "@channel"

            elif obj_id == "everyone":
                replacement = "@everyone"

            elif match[0:5] == "!date":
                make_bold = False
                date_parts = match.split("^")
                if len(date_parts) > 1:
                    replacement = self._locale_helper.get_datetime_formatted_str(
                        date_parts[1]
                    )
                else:
                    replacement = "(failed to parse date)"

            else:
                replacement = f"@special_{obj_id}"

        else:
            # match is an URL
            link_parts = match.split("|")
            if len(link_parts) == 2:
                url = link_parts[0]
                text = link_parts[1]
            else:
                url = match
                text = match

            make_bold = False
            replacement = f'<a href="{url}">{text}</a>'

        if make_bold:
            replacement = f"<b>{replacement}</b>"

        return replacement



# -*- coding: latin-1 -*-

"HTML Renderer for FPDF.py"

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2010 Mariano Reingart"
__license__ = "LGPL 3.0"

# Inspired by tuto5.py and several examples from fpdf.org, html2fpdf, etc.

from .fpdf import FPDF
from .py3k import PY3K, basestring, unicode, HTMLParser

DEBUG = False


def px2mm(px):
    return int(px) * 25.4 / 72.0


def hex2dec(color="#000000"):
    if color:
        r = int(color[1:3], 16)
        g = int(color[3:5], 16)
        b = int(color[5:7], 16)
        return r, g, b


class HTML2FPDF(HTMLParser):
    "Render basic HTML to FPDF"

    def __init__(self, pdf, image_map=None):
        HTMLParser.__init__(self)
        self.style = {}
        self.pre = False
        self.href = ""
        self.align = ""
        self.page_links = {}
        self.font = None
        self.font_stack = []
        self.pdf = pdf
        self.image_map = image_map or (lambda src: src)
        self.r = self.g = self.b = 0
        self.indent = 0
        self.bullet = []
        self.set_font("times", 12)
        self.font_face = "times"  # initialize font
        self.color = 0  # initialize font color
        self.table = None  # table attributes
        self.table_col_width = None  # column (header) widths
        self.table_col_index = None  # current column index
        self.td = None  # cell attributes
        self.th = False  # header enabled
        self.tr = None
        self.theader = None  # table header cells
        self.tfooter = None  # table footer cells
        self.thead = None
        self.tfoot = None
        self.theader_out = self.tfooter_out = False
        self.hsize = dict(h1=2, h2=1.5, h3=1.17, h4=1, h5=0.83, h6=0.67)

    def width2mm(self, length):
        if length[-1] == "%":
            total = self.pdf.w - self.pdf.r_margin - self.pdf.l_margin
            if self.table["width"][-1] == "%":
                total *= int(self.table["width"][:-1]) / 100.0
            return int(length[:-1]) * total / 101.0
        else:
            return int(length) / 6.0

    def handle_data(self, txt):
        if self.td is not None:  # drawing a table?
            if "width" not in self.td and "colspan" not in self.td:
                try:
                    l = [self.table_col_width[self.table_col_index]]
                except IndexError:
                    raise RuntimeError(
                        "Table column/cell width not specified, unable to continue"
                    )
            elif "colspan" in self.td:
                i = self.table_col_index
                colspan = int(self.td["colspan"])
                l = self.table_col_width[i : i + colspan]
            else:
                l = [self.td.get("width", "240")]
            w = sum([self.width2mm(length) for length in l])
            h = int(self.td.get("height", 0)) // 4 or self.h * 1.30
            self.table_h = h
            border = int(self.table.get("border", 0))
            if not self.th:
                align = self.td.get("align", "L")[0].upper()
                border = border and "LR"
            else:
                self.set_style("B", True)
                border = border or "B"
                align = self.td.get("align", "C")[0].upper()
            bgcolor = hex2dec(self.td.get("bgcolor", self.tr.get("bgcolor", "")))
            # parsing table header/footer (drawn later):
            if self.thead is not None:
                self.theader.append(((w, h, txt, border, 0, align), bgcolor))
            if self.tfoot is not None:
                self.tfooter.append(((w, h, txt, border, 0, align), bgcolor))
            # check if reached end of page, add table footer and header:
            height = h + (self.tfooter and self.tfooter[0][0][1] or 0)
            if self.pdf.y + height > self.pdf.page_break_trigger and not self.th:
                self.output_table_footer()
                self.pdf.add_page(same=True)
                self.theader_out = self.tfooter_out = False
            if self.tfoot is None and self.thead is None:
                if not self.theader_out:
                    self.output_table_header()
                self.box_shadow(w, h, bgcolor)
                if DEBUG:
                    print("td cell", self.pdf.x, w, txt, "*")
                self.pdf.cell(w, h, txt, border, 0, align)
        elif self.table is not None:
            # ignore anything else than td inside a table
            pass
        elif self.align:
            if DEBUG:
                print("cell", txt, "*")
            self.pdf.cell(0, self.h, txt, 0, 1, self.align[0].upper(), self.href)
        else:
            txt = txt.replace("\n", " ")
            if self.href:
                self.put_link(self.href, txt)
            else:
                if DEBUG:
                    print("write", txt, "*")
                self.pdf.write(self.h, txt)

    def box_shadow(self, w, h, bgcolor):
        if DEBUG:
            print("box_shadow", w, h, bgcolor)
        if bgcolor:
            fill_color = self.pdf.fill_color
            self.pdf.set_fill_color(*bgcolor)
            self.pdf.rect(self.pdf.x, self.pdf.y, w, h, "F")
            self.pdf.fill_color = fill_color

    def output_table_header(self):
        if self.theader:
            b = self.b
            x = self.pdf.x
            self.pdf.set_x(self.table_offset)
            self.set_style("B", True)
            for cell, bgcolor in self.theader:
                self.box_shadow(cell[0], cell[1], bgcolor)
                self.pdf.cell(*cell)
            self.set_style("B", b)
            self.pdf.ln(self.theader[0][0][1])
            self.pdf.set_x(self.table_offset)
            # self.pdf.set_x(x)
        self.theader_out = True

    def output_table_footer(self):
        if self.tfooter:
            x = self.pdf.x
            self.pdf.set_x(self.table_offset)
            # TODO: self.output_table_sep()
            for cell, bgcolor in self.tfooter:
                self.box_shadow(cell[0], cell[1], bgcolor)
                self.pdf.cell(*cell)
            self.pdf.ln(self.tfooter[0][0][1])
            self.pdf.set_x(x)
        if int(self.table.get("border", 0)):
            self.output_table_sep()
        self.tfooter_out = True

    def output_table_sep(self):
        self.pdf.set_x(self.table_offset)
        x1 = self.pdf.x
        y1 = self.pdf.y
        w = sum([self.width2mm(lenght) for lenght in self.table_col_width])
        self.pdf.line(x1, y1, x1 + w, y1)

    def handle_starttag(self, tag, attrs):
        attrs = dict(attrs)
        if DEBUG:
            print("STARTTAG", tag, attrs)
        if tag == "b" or tag == "i" or tag == "u":
            self.set_style(tag, 1)
        if tag == "a":
            self.href = attrs["href"]
        if tag == "br":
            self.pdf.ln(5)
        if tag == "p":
            self.pdf.ln(5)
            if attrs:
                if attrs:
                    self.align = attrs.get("align")
        if tag in self.hsize:
            k = self.hsize[tag]
            self.pdf.ln(5 * k)
            self.pdf.set_text_color(150, 0, 0)
            self.pdf.set_font_size(12 * k)
            if attrs:
                self.align = attrs.get("align")
        if tag == "hr":
            self.put_line()
        if tag == "pre":
            self.pdf.set_font("Courier", "", 11)
            self.pdf.set_font_size(11)
            self.set_style("B", False)
            self.set_style("I", False)
            self.pre = True
        if tag == "blockquote":
            self.set_text_color(100, 0, 45)
            self.pdf.ln(3)
        if tag == "ul":
            self.indent += 1
            self.bullet.append("\x95")
        if tag == "ol":
            self.indent += 1
            self.bullet.append(0)
        if tag == "li":
            self.pdf.ln(self.h + 2)
            self.pdf.set_text_color(190, 0, 0)
            bullet = self.bullet[self.indent - 1]
            if not isinstance(bullet, basestring):
                bullet += 1
                self.bullet[self.indent - 1] = bullet
                bullet = "%s. " % bullet
            self.pdf.write(self.h, "%s%s " % (" " * 5 * self.indent, bullet))
            self.set_text_color()
        if tag == "font":
            # save previous font state:
            self.font_stack.append((self.font_face, self.font_size, self.color))
            if "color" in attrs:
                color = hex2dec(attrs["color"])
                self.set_text_color(*color)
                self.color = color
            if "face" in attrs:
                face = attrs.get("face").lower()
                try:
                    self.pdf.set_font(face)
                    self.font_face = face
                except RuntimeError:
                    pass  # font not found, ignore
            if "size" in attrs:
                size = int(attrs.get("size"))
                self.pdf.set_font(self.font_face, size=int(size))
                self.font_size = size
        if tag == "table":
            self.table = dict([(k.lower(), v) for k, v in attrs.items()])
            if not "width" in self.table:
                self.table["width"] = "100%"
            if self.table["width"][-1] == "%":
                w = self.pdf.w - self.pdf.r_margin - self.pdf.l_margin
                w *= int(self.table["width"][:-1]) / 100.0
                self.table_offset = (self.pdf.w - w) / 2.0
            self.table_col_width = []
            self.theader_out = self.tfooter_out = False
            self.theader = []
            self.tfooter = []
            self.thead = None
            self.tfoot = None
            self.table_h = 0
            self.pdf.ln()
        if tag == "tr":
            self.tr = dict([(k.lower(), v) for k, v in attrs.items()])
            self.table_col_index = 0
            self.pdf.set_x(self.table_offset)
        if tag == "td":
            self.td = dict([(k.lower(), v) for k, v in attrs.items()])
        if tag == "th":
            self.td = dict([(k.lower(), v) for k, v in attrs.items()])
            self.th = True
            if "width" in self.td:
                self.table_col_width.append(self.td["width"])
        if tag == "thead":
            self.thead = {}
        if tag == "tfoot":
            self.tfoot = {}
        if tag == "img":
            if "src" in attrs:
                x = self.pdf.get_x()
                y = self.pdf.get_y()
                w = px2mm(attrs.get("width", 0))
                h = px2mm(attrs.get("height", 0))
                if self.align and self.align[0].upper() == "C":
                    x = (self.pdf.w - x) / 2.0 - w / 2.0
                self.pdf.image(self.image_map(attrs["src"]), x, y, w, h, link=self.href)
                self.pdf.set_x(x + w)
                self.pdf.set_y(y + h)
        if tag == "b" or tag == "i" or tag == "u":
            self.set_style(tag, True)
        if tag == "center":
            self.align = "Center"

    def handle_endtag(self, tag):
        # Closing tag
        if DEBUG:
            print("ENDTAG", tag)
        if tag == "h1" or tag == "h2" or tag == "h3" or tag == "h4":
            self.pdf.ln(6)
            self.set_font()
            self.set_style()
            self.align = None
        if tag == "pre":
            self.pdf.set_font(self.font or "Times", "", 12)
            self.pdf.set_font_size(12)
            self.pre = False
        if tag == "blockquote":
            self.set_text_color(0, 0, 0)
            self.pdf.ln(3)
        if tag == "strong":
            tag = "b"
        if tag == "em":
            tag = "i"
        if tag == "b" or tag == "i" or tag == "u":
            self.set_style(tag, False)
        if tag == "a":
            self.href = ""
        if tag == "p":
            self.align = ""
        if tag in ("ul", "ol"):
            self.indent -= 1
            self.bullet.pop()
        if tag == "table":
            if not self.tfooter_out:
                self.output_table_footer()
            self.table = None
            self.th = False
            self.theader = None
            self.tfooter = None
            self.pdf.ln()
        if tag == "thead":
            self.thead = None
        if tag == "tfoot":
            self.tfoot = None
        if tag == "tbody":
            # draw a line separator between table bodies
            self.pdf.set_x(self.table_offset)
            self.output_table_sep()
        if tag == "tr":
            h = self.table_h
            if self.tfoot is None:
                self.pdf.ln(h)
            self.tr = None
        if tag == "td" or tag == "th":
            if self.th:
                if DEBUG:
                    print("revert style")
                self.set_style("B", False)  # revert style
            self.table_col_index += int(self.td.get("colspan", "1"))
            self.td = None
            self.th = False
        if tag == "font":
            # recover last font state
            face, size, color = self.font_stack.pop()
            if face:
                self.pdf.set_text_color(0, 0, 0)
                self.color = None
            self.set_font(face, size)
            self.font = None
        if tag == "center":
            self.align = None

    def set_font(self, face=None, size=None):
        if face:
            self.font_face = face
        if size:
            self.font_size = size
            self.h = size / 72.0 * 25.4
            if DEBUG:
                print("H", self.h)
        self.pdf.set_font(self.font_face or "times", "", 12)
        self.pdf.set_font_size(self.font_size or 12)
        self.set_style("u", False)
        self.set_style("b", False)
        self.set_style("i", False)
        self.set_text_color()

    def set_style(self, tag=None, enable=None):
        # Modify style and select corresponding font
        if tag:
            t = self.style.get(tag.lower())
            self.style[tag.lower()] = enable
        style = ""
        for s in ("b", "i", "u"):
            if self.style.get(s):
                style += s
        if DEBUG:
            print("SET_FONT_STYLE", style)
        self.pdf.set_font("", style)

    def set_text_color(self, r=None, g=0, b=0):
        if r is None:
            self.pdf.set_text_color(self.r, self.g, self.b)
        else:
            self.pdf.set_text_color(r, g, b)
            self.r = r
            self.g = g
            self.b = b

    def put_link(self, url, txt):
        # Put a hyperlink
        self.set_text_color(0, 0, 255)
        self.set_style("u", True)
        self.pdf.write(5, txt, url)
        self.set_style("u", False)
        self.set_text_color(0)

    def put_line(self):
        self.pdf.ln(2)
        self.pdf.line(
            self.pdf.get_x(), self.pdf.get_y(), self.pdf.get_x() + 187, self.pdf.get_y()
        )
        self.pdf.ln(3)


class HTMLMixin(object):
    def write_html(self, text, image_map=None):
        "Parse HTML and convert it to PDF"
        h2p = HTML2FPDF(self, image_map)
        text = h2p.unescape(text)  # To deal with HTML entities
        h2p.feed(text)



import logging
import html
import json
from pathlib import Path


logger = logging.getLogger(__name__)


def transform_encoding(text: str) -> str:
    """adjust encoding to latin-1 and transform HTML entities"""
    text2 = html.unescape(text)
    text2 = text2.encode("utf-8", "replace").decode("utf-8")
    text2 = text2.replace("\t", "    ")
    return text2


def read_array_from_json_file(filepath: Path, quiet=False) -> list:
    """reads a json file and returns its contents as array"""
    my_file = filepath.parent / (filepath.name + ".json")
    if not my_file.is_file:
        if quiet is False:
            logger.warning("file does not exist: %s", filepath)
        arr = list()
    else:
        try:
            with my_file.open("r", encoding="utf-8") as f:
                arr = json.load(f)
        except IOError:
            if quiet is False:
                logger.warning("failed to read from %s: ", my_file, exc_info=True)
            arr = list()

    return arr


def write_array_to_json_file(arr, filepath: Path) -> None:
    """writes array to a json file"""
    my_file = filepath.parent / (filepath.name + ".json")
    logger.info("Writing file: %s", filepath)
    try:
        with my_file.open("w", encoding="utf-8") as f:
            json.dump(arr, f, sort_keys=True, indent=4, ensure_ascii=False)
    except IOError:
        logger.error("failed to write to %s", my_file, exc_info=True)



"""Defines all global settings incl. from configuration files"""

from ast import literal_eval
import configparser
from pathlib import Path

_FILE_NAME_BASE = "slackchannel2pdf"
_CONF_FILE_NAME = f"{_FILE_NAME_BASE}.ini"
_LOG_FILE_NAME = f"{_FILE_NAME_BASE}.log"

_DEFAULTS_PATH = Path(__file__).parent


def _configparser_convert_str(x):
    result = literal_eval(x)
    if not isinstance(result, str):
        raise configparser.ParsingError(f"Needs to be a string type: {x}")
    return result


def config_parser(
    defaults_path: Path, home_path: Path = None, cwd_path: Path = None
) -> configparser.ConfigParser:
    parser = configparser.ConfigParser(converters={"str": _configparser_convert_str})
    config_file_paths = [defaults_path / _CONF_FILE_NAME]
    if home_path:
        config_file_paths.append(home_path / _CONF_FILE_NAME)
    if cwd_path:
        config_file_paths.append(cwd_path / _CONF_FILE_NAME)
    found = parser.read(config_file_paths)
    if not found:
        raise RuntimeError("Can not find a configuration file anywhere")
    return parser


_my_config = config_parser(
    defaults_path=_DEFAULTS_PATH, home_path=Path.home(), cwd_path=Path.cwd()
)

# style and layout settings for PDF
PAGE_UNITS_DEFAULT = "mm"
FONT_FAMILY_DEFAULT = "NotoSans"
FONT_FAMILY_MONO_DEFAULT = "NotoSansMono"

PAGE_ORIENTATION_DEFAULT = _my_config.getstr("pdf", "page_orientation")
PAGE_FORMAT_DEFAULT = _my_config.getstr("pdf", "page_format")
FONT_SIZE_NORMAL = _my_config.getint("pdf", "font_size_normal")
FONT_SIZE_LARGE = _my_config.getint("pdf", "font_size_large")
FONT_SIZE_SMALL = _my_config.getint("pdf", "font_size_small")
LINE_HEIGHT_DEFAULT = _my_config.getint("pdf", "line_height_default")
LINE_HEIGHT_SMALL = _my_config.getint("pdf", "line_height_small")
MARGIN_LEFT = _my_config.getint("pdf", "margin_left")
TAB_WIDTH = _my_config.getint("pdf", "tab_width")

# locale
FALLBACK_LOCALE = _my_config.getstr("locale", "fallback_locale")

# slack
MINUTES_UNTIL_USERNAME_REPEATS = _my_config.getint(
    "slack", "minutes_until_username_repeats"
)
MAX_MESSAGES_PER_CHANNEL = _my_config.getint("slack", "max_messages_per_channel")
SLACK_PAGE_LIMIT = _my_config.getint("slack", "slack_page_limit")


def setup_logging(config: configparser.ConfigParser) -> None:
    config_logging = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "console": {"format": "[%(levelname)s] %(message)s"},
            "file": {"format": "%(asctime)s [%(levelname)s] %(name)s: %(message)s"},
        },
        "handlers": {
            "console": {
                "level": config.getstr("logging", "console_log_level"),
                "formatter": "console",
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",  # Default is stderr
            }
        },
        "loggers": {
            "": {  # root logger
                "handlers": ["console"],
                "level": "DEBUG",
                "propagate": False,
            },
        },
    }

    # add log file if configured
    log_file_enabled = config.getboolean("logging", "log_file_enabled", fallback=False)
    if log_file_enabled:
        file_log_path_full = config.getstr("logging", "log_file_path", fallback=None)
        filename = (
            Path(file_log_path_full) / _LOG_FILE_NAME
            if file_log_path_full
            else _LOG_FILE_NAME
        )
        config_logging["handlers"]["file"] = {
            "level": config.getstr("logging", "file_log_level"),
            "formatter": "file",
            "class": "logging.FileHandler",
            "filename": filename,
            "mode": "a",
        }
        config_logging["loggers"][""]["handlers"].append("file")
    return config_logging


DEFAULT_LOGGING = setup_logging(_my_config)



import datetime as dt
import logging

from babel import Locale, UnknownLocaleError
from babel.dates import format_datetime, format_time, format_date
import pytz
from tzlocal import get_localzone

from . import settings


logger = logging.getLogger(__name__)


class LocaleHelper:
    """Helpers for converting date & time according to current locale and timezone"""

    def __init__(
        self,
        my_locale: Locale = None,
        my_tz: pytz.BaseTzInfo = None,
        author_info: dict = None,
    ) -> None:
        """
        Args:
        - my_locale: Primary locale to use
        - my_tz: Primary timezone to use
        - author_info: locale and timezone to use from this Slack response
        if my_locale and/or my_tz are not given
        """
        self._locale = self._determine_locale(my_locale, author_info)
        self._timezone = self._determine_timezone(my_tz, author_info)

    @staticmethod
    def _determine_locale(my_locale: Locale = None, author_info: dict = None) -> Locale:
        if my_locale:
            if not isinstance(my_locale, Locale):
                raise TypeError("my_locale must be a babel Locale object")
        else:
            if author_info:
                try:
                    my_locale = Locale.parse(author_info["locale"], sep="-")
                except UnknownLocaleError:
                    logger.warning("Could not use locale info from Slack")
                    my_locale = Locale.default()
            else:
                my_locale = Locale.default()
        if not my_locale:
            my_locale = Locale.parse(settings.FALLBACK_LOCALE)
        return my_locale

    @staticmethod
    def _determine_timezone(
        my_tz: pytz.BaseTzInfo = None, author_info: dict = None
    ) -> pytz.BaseTzInfo:
        if my_tz:
            if not isinstance(my_tz, pytz.BaseTzInfo):
                raise TypeError("my_tz must be of type pytz")
        else:
            if author_info:
                try:
                    my_tz = pytz.timezone(author_info["tz"])
                except pytz.exceptions.UnknownTimeZoneError:
                    logger.warning("Could not use timezone info from Slack")
                    my_tz = get_localzone()
            else:
                my_tz = get_localzone()
        if not my_tz:
            my_tz = pytz.UTC
        return my_tz

    @property
    def locale(self) -> Locale:
        return self._locale

    @property
    def timezone(self) -> pytz.BaseTzInfo:
        return self._timezone

    def format_date_full_str(self, my_datetime: dt.datetime) -> str:
        return format_date(my_datetime, format="full", locale=self.locale)

    def format_datetime_str(self, my_datetime: dt.datetime) -> str:
        """returns formated datetime string for given dt using locale"""
        return format_datetime(my_datetime, format="short", locale=self.locale)

    def get_datetime_formatted_str(self, ts: int) -> str:
        """return given timestamp as formated datetime string using locale"""
        my_datetime = self.get_datetime_from_ts(ts)
        return format_datetime(my_datetime, format="short", locale=self.locale)

    def get_time_formatted_str(self, ts: int) -> str:
        """return given timestamp as formated datetime string using locale"""
        my_datetime = self.get_datetime_from_ts(ts)
        return format_time(my_datetime, format="short", locale=self.locale)

    def get_datetime_from_ts(self, ts: int) -> dt.datetime:
        """returns datetime object of a unix timestamp with local timezone"""
        my_datetime = dt.datetime.fromtimestamp(float(ts), pytz.UTC)
        return my_datetime.astimezone(self.timezone)



#!/usr/bin/env python
# -*- coding: latin-1 -*-
# ****************************************************************************
# * Software: FPDF for python                                                *
# * Version:  1.7.1                                                          *
# * Date:     2010-09-10                                                     *
# * Last update: 2012-08-16                                                  *
# * License:  LGPL v3.0                                                      *
# *                                                                          *
# * Original Author (PHP):  Olivier PLATHEY 2004-12-31                       *
# * Ported to Python 2.4 by Max (maxpat78@yahoo.it) on 2006-05               *
# * Maintainer:  Mariano Reingart (reingart@gmail.com) et al since 2008 est. *
# * NOTE: 'I' and 'D' destinations are disabled, and simply print to STDOUT  *
# ****************************************************************************

from __future__ import division, with_statement

from datetime import datetime
from functools import wraps
import math
import errno
import os, sys, zlib, struct, re, tempfile, struct

from .ttfonts import TTFontFile
from .fonts import fpdf_charwidths
from .php import substr, sprintf, print_r, UTF8ToUTF16BE, UTF8StringToArray
from .py3k import (
    PY3K,
    pickle,
    urlopen,
    BytesIO,
    Image,
    basestring,
    unicode,
    exception,
    b,
    hashpath,
)

# Global variables
FPDF_VERSION = "1.7.2a"  # with Kalkoken patch
FPDF_FONT_DIR = os.path.join(os.path.dirname(__file__), "font")
FPDF_CACHE_MODE = 0  # 0 - in same folder, 1 - none, 2 - hash
FPDF_CACHE_DIR = None
SYSTEM_TTFONTS = None

PAGE_FORMATS = {
    "a3": (841.89, 1190.55),
    "a4": (595.28, 841.89),
    "a5": (420.94, 595.28),
    "letter": (612, 792),
    "legal": (612, 1008),
}


def set_global(var, val):
    globals()[var] = val


def load_cache(filename):
    """Return unpickled object, or None if cache unavailable"""
    if not filename:
        return None
    try:
        with open(filename, "rb") as fh:
            return pickle.load(fh)
    except (IOError, ValueError):  # File missing, unsupported pickle, etc
        return None


class FPDF(object):
    "PDF Generation class"

    def __init__(self, orientation="P", unit="mm", format="A4"):
        # Some checks
        self._dochecks()
        # Initialization of properties
        self.offsets = {}  # array of object offsets
        self.page = 0  # current page number
        self.n = 2  # current object number
        self.buffer = ""  # buffer holding in-memory PDF
        self.pages = {}  # array containing pages and metadata
        self.state = 0  # current document state
        self.fonts = {}  # array of used fonts
        self.font_files = {}  # array of font files
        self.diffs = {}  # array of encoding differences
        self.images = {}  # array of used images
        self.page_links = {}  # array of links in pages
        self.links = {}  # array of internal links
        self.in_footer = 0  # flag set when processing footer
        self.lastw = 0
        self.lasth = 0  # height of last cell printed
        self.font_family = ""  # current font family
        self.font_style = ""  # current font style
        self.font_size_pt = 12  # current font size in points
        self.font_stretching = 100  # current font stretching
        self.underline = 0  # underlining flag
        self.draw_color = "0 G"
        self.fill_color = "0 g"
        self.text_color = "0 g"
        self.color_flag = 0  # indicates whether fill and text colors are different
        self.ws = 0  # word spacing
        self.angle = 0
        # Standard fonts
        self.core_fonts = {
            "courier": "Courier",
            "courierB": "Courier-Bold",
            "courierI": "Courier-Oblique",
            "courierBI": "Courier-BoldOblique",
            "helvetica": "Helvetica",
            "helveticaB": "Helvetica-Bold",
            "helveticaI": "Helvetica-Oblique",
            "helveticaBI": "Helvetica-BoldOblique",
            "times": "Times-Roman",
            "timesB": "Times-Bold",
            "timesI": "Times-Italic",
            "timesBI": "Times-BoldItalic",
            "symbol": "Symbol",
            "zapfdingbats": "ZapfDingbats",
        }
        self.core_fonts_encoding = "latin-1"
        # Scale factor
        if unit == "pt":
            self.k = 1
        elif unit == "mm":
            self.k = 72 / 25.4
        elif unit == "cm":
            self.k = 72 / 2.54
        elif unit == "in":
            self.k = 72.0
        else:
            self.error("Incorrect unit: " + unit)
        # Page format
        self.fw_pt, self.fh_pt = self.get_page_format(format, self.k)
        self.dw_pt = self.fw_pt
        self.dh_pt = self.fh_pt
        self.fw = self.fw_pt / self.k
        self.fh = self.fh_pt / self.k
        # Page orientation
        orientation = orientation.lower()
        if orientation in ("p", "portrait"):
            self.def_orientation = "P"
            self.w_pt = self.fw_pt
            self.h_pt = self.fh_pt
        elif orientation in ("l", "landscape"):
            self.def_orientation = "L"
            self.w_pt = self.fh_pt
            self.h_pt = self.fw_pt
        else:
            self.error("Incorrect orientation: " + orientation)
        self.cur_orientation = self.def_orientation
        self.w = self.w_pt / self.k
        self.h = self.h_pt / self.k
        # Page margins (1 cm)
        margin = 28.35 / self.k
        self.set_margins(margin, margin)
        # Interior cell margin (1 mm)
        self.c_margin = margin / 10.0
        # line width (0.2 mm)
        self.line_width = 0.567 / self.k
        # Automatic page break
        self.set_auto_page_break(1, 2 * margin)
        # Full width display mode
        self.set_display_mode("fullwidth")
        # Enable compression
        self.set_compression(1)
        # Set default PDF version number
        self.pdf_version = "1.3"

    @staticmethod
    def get_page_format(format, k):
        "Return scale factor, page w and h size in points"
        if isinstance(format, basestring):
            format = format.lower()
            if format in PAGE_FORMATS:
                return PAGE_FORMATS[format]
            else:
                raise RuntimeError("Unknown page format: " + format)
        else:
            return (format[0] * k, format[1] * k)

    def check_page(fn):
        "Decorator to protect drawing methods"

        @wraps(fn)
        def wrapper(self, *args, **kwargs):
            if not self.page and not kwargs.get("split_only"):
                self.error("No page open, you need to call add_page() first")
            else:
                return fn(self, *args, **kwargs)

        return wrapper

    def set_margins(self, left, top, right=-1):
        "Set left, top and right margins"
        self.l_margin = left
        self.t_margin = top
        if right == -1:
            right = left
        self.r_margin = right

    def set_left_margin(self, margin):
        "Set left margin"
        self.l_margin = margin
        if self.page > 0 and self.x < margin:
            self.x = margin

    def set_top_margin(self, margin):
        "Set top margin"
        self.t_margin = margin

    def set_right_margin(self, margin):
        "Set right margin"
        self.r_margin = margin

    def set_auto_page_break(self, auto, margin=0):
        "Set auto page break mode and triggering margin"
        self.auto_page_break = auto
        self.b_margin = margin
        self.page_break_trigger = self.h - margin

    def set_display_mode(self, zoom, layout="continuous"):
        """Set display mode in viewer

        The "zoom" argument may be 'fullpage', 'fullwidth', 'real',
        'default', or a number, interpreted as a percentage."""

        if (
            zoom == "fullpage"
            or zoom == "fullwidth"
            or zoom == "real"
            or zoom == "default"
            or not isinstance(zoom, basestring)
        ):
            self.zoom_mode = zoom
        else:
            self.error("Incorrect zoom display mode: " + zoom)
        if (
            layout == "single"
            or layout == "continuous"
            or layout == "two"
            or layout == "default"
        ):
            self.layout_mode = layout
        else:
            self.error("Incorrect layout display mode: " + layout)

    def set_compression(self, compress):
        "Set page compression"
        self.compress = compress

    def set_title(self, title):
        "Title of document"
        self.title = title

    def set_subject(self, subject):
        "Subject of document"
        self.subject = subject

    def set_author(self, author):
        "Author of document"
        self.author = author

    def set_keywords(self, keywords):
        "Keywords of document"
        self.keywords = keywords

    def set_creator(self, creator):
        "Creator of document"
        self.creator = creator

    def set_doc_option(self, opt, value):
        "Set document option"
        if opt == "core_fonts_encoding":
            self.core_fonts_encoding = value
        else:
            self.error('Unknown document option "%s"' % str(opt))

    def alias_nb_pages(self, alias="{nb}"):
        "Define an alias for total number of pages"
        self.str_alias_nb_pages = alias
        return alias

    def error(self, msg):
        "Fatal error"
        raise RuntimeError("FPDF error: " + msg)

    def open(self):
        "Begin document"
        self.state = 1

    def close(self):
        "Terminate document"
        if self.state == 3:
            return
        if self.page == 0:
            self.add_page()
        # Page footer
        self.in_footer = 1
        self.footer()
        self.in_footer = 0
        # close page
        self._endpage()
        # close document
        self._enddoc()

    def add_page(self, orientation="", format="", same=False):
        "Start a new page, if same page format will be same as previous"
        if self.state == 0:
            self.open()
        family = self.font_family
        if self.underline:
            style = self.font_style + "U"
        else:
            style = self.font_style
        size = self.font_size_pt
        lw = self.line_width
        dc = self.draw_color
        fc = self.fill_color
        tc = self.text_color
        cf = self.color_flag
        stretching = self.font_stretching
        if self.page > 0:
            # Page footer
            self.in_footer = 1
            self.footer()
            self.in_footer = 0
            # close page
            self._endpage()
        # Start new page
        self._beginpage(orientation, format, same)
        # Set line cap style to square
        self._out("2 J")
        # Set line width
        self.line_width = lw
        self._out(sprintf("%.2f w", lw * self.k))
        # Set font
        if family:
            self.set_font(family, style, size)
        # Set colors
        self.draw_color = dc
        if dc != "0 G":
            self._out(dc)
        self.fill_color = fc
        if fc != "0 g":
            self._out(fc)
        self.text_color = tc
        self.color_flag = cf
        # Page header
        self.header()
        # Restore line width
        if self.line_width != lw:
            self.line_width = lw
            self._out(sprintf("%.2f w", lw * self.k))
        # Restore font
        if family:
            self.set_font(family, style, size)
        # Restore colors
        if self.draw_color != dc:
            self.draw_color = dc
            self._out(dc)
        if self.fill_color != fc:
            self.fill_color = fc
            self._out(fc)
        self.text_color = tc
        self.color_flag = cf
        # Restore stretching
        if stretching != 100:
            self.set_stretching(stretching)

    def header(self):
        "Header to be implemented in your own inherited class"
        pass

    def footer(self):
        "Footer to be implemented in your own inherited class"
        pass

    def page_no(self):
        "Get current page number"
        return self.page

    def set_draw_color(self, r, g=-1, b=-1):
        "Set color for all stroking operations"
        if (r == 0 and g == 0 and b == 0) or g == -1:
            self.draw_color = sprintf("%.3f G", r / 255.0)
        else:
            self.draw_color = sprintf(
                "%.3f %.3f %.3f RG", r / 255.0, g / 255.0, b / 255.0
            )
        if self.page > 0:
            self._out(self.draw_color)

    def set_fill_color(self, r, g=-1, b=-1):
        "Set color for all filling operations"
        if (r == 0 and g == 0 and b == 0) or g == -1:
            self.fill_color = sprintf("%.3f g", r / 255.0)
        else:
            self.fill_color = sprintf(
                "%.3f %.3f %.3f rg", r / 255.0, g / 255.0, b / 255.0
            )
        self.color_flag = self.fill_color != self.text_color
        if self.page > 0:
            self._out(self.fill_color)

    def set_text_color(self, r, g=-1, b=-1):
        "Set color for text"
        if (r == 0 and g == 0 and b == 0) or g == -1:
            self.text_color = sprintf("%.3f g", r / 255.0)
        else:
            self.text_color = sprintf(
                "%.3f %.3f %.3f rg", r / 255.0, g / 255.0, b / 255.0
            )
        self.color_flag = self.fill_color != self.text_color

    def get_string_width(self, s, normalized=False):
        "Get width of a string in the current font"
        # normalized is parameter for internal use
        s = s if normalized else self.normalize_text(s)
        cw = self.current_font["cw"]
        w = 0
        l = len(s)
        if self.unifontsubset:
            for char in s:
                char = ord(char)
                if len(cw) > char:
                    w += cw[char]  # ord(cw[2*char])<<8 + ord(cw[2*char+1])
                # elif (char>0 and char<128 and isset($cw[chr($char)])) { $w += $cw[chr($char)]; }
                elif self.current_font["desc"]["MissingWidth"]:
                    w += self.current_font["desc"]["MissingWidth"]
                # elif (isset($this->CurrentFont['MissingWidth'])) { $w += $this->CurrentFont['MissingWidth']; }
                else:
                    w += 500
        else:
            for i in range(0, l):
                w += cw.get(s[i], 0)
        if self.font_stretching != 100:
            w = w * self.font_stretching / 100.0
        return w * self.font_size / 1000.0

    def set_line_width(self, width):
        "Set line width"
        self.line_width = width
        if self.page > 0:
            self._out(sprintf("%.2f w", width * self.k))

    @check_page
    def line(self, x1, y1, x2, y2):
        "Draw a line"
        self._out(
            sprintf(
                "%.2f %.2f m %.2f %.2f l S",
                x1 * self.k,
                (self.h - y1) * self.k,
                x2 * self.k,
                (self.h - y2) * self.k,
            )
        )

    def _set_dash(self, dash_length=False, space_length=False):
        if dash_length and space_length:
            s = sprintf("[%.3f %.3f] 0 d", dash_length * self.k, space_length * self.k)
        else:
            s = "[] 0 d"
        self._out(s)

    @check_page
    def dashed_line(self, x1, y1, x2, y2, dash_length=1, space_length=1):
        """Draw a dashed line. Same interface as line() except:
        - dash_length: Length of the dash
        - space_length: Length of the space between dashes"""
        self._set_dash(dash_length, space_length)
        self.line(x1, y1, x2, y2)
        self._set_dash()

    @check_page
    def rect(self, x, y, w, h, style=""):
        "Draw a rectangle"
        if style == "F":
            op = "f"
        elif style == "FD" or style == "DF":
            op = "B"
        else:
            op = "S"
        self._out(
            sprintf(
                "%.2f %.2f %.2f %.2f re %s",
                x * self.k,
                (self.h - y) * self.k,
                w * self.k,
                -h * self.k,
                op,
            )
        )

    @check_page
    def ellipse(self, x, y, w, h, style=""):
        "Draw a ellipse"
        if style == "F":
            op = "f"
        elif style == "FD" or style == "DF":
            op = "B"
        else:
            op = "S"

        cx = x + w / 2.0
        cy = y + h / 2.0
        rx = w / 2.0
        ry = h / 2.0

        lx = 4.0 / 3.0 * (math.sqrt(2) - 1) * rx
        ly = 4.0 / 3.0 * (math.sqrt(2) - 1) * ry

        self._out(
            sprintf(
                "%.2f %.2f m %.2f %.2f %.2f %.2f %.2f %.2f c",
                (cx + rx) * self.k,
                (self.h - cy) * self.k,
                (cx + rx) * self.k,
                (self.h - (cy - ly)) * self.k,
                (cx + lx) * self.k,
                (self.h - (cy - ry)) * self.k,
                cx * self.k,
                (self.h - (cy - ry)) * self.k,
            )
        )
        self._out(
            sprintf(
                "%.2f %.2f %.2f %.2f %.2f %.2f c",
                (cx - lx) * self.k,
                (self.h - (cy - ry)) * self.k,
                (cx - rx) * self.k,
                (self.h - (cy - ly)) * self.k,
                (cx - rx) * self.k,
                (self.h - cy) * self.k,
            )
        )
        self._out(
            sprintf(
                "%.2f %.2f %.2f %.2f %.2f %.2f c",
                (cx - rx) * self.k,
                (self.h - (cy + ly)) * self.k,
                (cx - lx) * self.k,
                (self.h - (cy + ry)) * self.k,
                cx * self.k,
                (self.h - (cy + ry)) * self.k,
            )
        )
        self._out(
            sprintf(
                "%.2f %.2f %.2f %.2f %.2f %.2f c %s",
                (cx + lx) * self.k,
                (self.h - (cy + ry)) * self.k,
                (cx + rx) * self.k,
                (self.h - (cy + ly)) * self.k,
                (cx + rx) * self.k,
                (self.h - cy) * self.k,
                op,
            )
        )

    def add_font(self, family, style="", fname="", uni=False):
        "Add a TrueType or Type1 font"
        family = family.lower()
        if fname == "":
            fname = family.replace(" ", "") + style.lower() + ".pkl"
        if family == "arial":
            family = "helvetica"
        style = style.upper()
        if style == "IB":
            style = "BI"
        fontkey = family + style
        if fontkey in self.fonts:
            # Font already added!
            return
        if uni:
            global SYSTEM_TTFONTS, FPDF_CACHE_MODE, FPDF_CACHE_DIR
            if os.path.exists(fname):
                ttffilename = fname
            elif FPDF_FONT_DIR and os.path.exists(os.path.join(FPDF_FONT_DIR, fname)):
                ttffilename = os.path.join(FPDF_FONT_DIR, fname)
            elif SYSTEM_TTFONTS and os.path.exists(os.path.join(SYSTEM_TTFONTS, fname)):
                ttffilename = os.path.join(SYSTEM_TTFONTS, fname)
            else:
                raise RuntimeError("TTF Font file not found: %s" % fname)
            name = ""
            if FPDF_CACHE_MODE == 0:
                unifilename = os.path.splitext(ttffilename)[0] + ".pkl"
            elif FPDF_CACHE_MODE == 2:
                unifilename = os.path.join(
                    FPDF_CACHE_DIR, hashpath(ttffilename) + ".pkl"
                )
            else:
                unifilename = None
            font_dict = load_cache(unifilename)
            if font_dict is None:
                ttf = TTFontFile()
                ttf.getMetrics(ttffilename)
                desc = {
                    "Ascent": int(round(ttf.ascent, 0)),
                    "Descent": int(round(ttf.descent, 0)),
                    "CapHeight": int(round(ttf.capHeight, 0)),
                    "Flags": ttf.flags,
                    "FontBBox": "[%s %s %s %s]"
                    % (
                        int(round(ttf.bbox[0], 0)),
                        int(round(ttf.bbox[1], 0)),
                        int(round(ttf.bbox[2], 0)),
                        int(round(ttf.bbox[3], 0)),
                    ),
                    "ItalicAngle": int(ttf.italicAngle),
                    "StemV": int(round(ttf.stemV, 0)),
                    "MissingWidth": int(round(ttf.defaultWidth, 0)),
                }
                # Generate metrics .pkl file
                font_dict = {
                    "name": re.sub("[ ()]", "", ttf.fullName),
                    "type": "TTF",
                    "desc": desc,
                    "up": round(ttf.underlinePosition),
                    "ut": round(ttf.underlineThickness),
                    "ttffile": ttffilename,
                    "fontkey": fontkey,
                    "originalsize": os.stat(ttffilename).st_size,
                    "cw": ttf.charWidths,
                }
                if unifilename:
                    try:
                        with open(unifilename, "wb") as fh:
                            pickle.dump(font_dict, fh)
                    except IOError:
                        if not exception().errno == errno.EACCES:
                            raise  # Not a permission error.
                del ttf
            if hasattr(self, "str_alias_nb_pages"):
                sbarr = list(range(0, 57))  # include numbers in the subset!
            else:
                sbarr = list(range(0, 32))
            self.fonts[fontkey] = {
                "i": len(self.fonts) + 1,
                "type": font_dict["type"],
                "name": font_dict["name"],
                "desc": font_dict["desc"],
                "up": font_dict["up"],
                "ut": font_dict["ut"],
                "cw": font_dict["cw"],
                "ttffile": font_dict["ttffile"],
                "fontkey": fontkey,
                "subset": sbarr,
                "unifilename": unifilename,
            }
            self.font_files[fontkey] = {
                "length1": font_dict["originalsize"],
                "type": "TTF",
                "ttffile": ttffilename,
            }
            self.font_files[fname] = {"type": "TTF"}
        else:
            with open(fname, "rb") as fontfile:
                font_dict = pickle.load(fontfile)
            self.fonts[fontkey] = {"i": len(self.fonts) + 1}
            self.fonts[fontkey].update(font_dict)
            diff = font_dict.get("diff")
            if diff:
                # Search existing encodings
                d = 0
                nb = len(self.diffs)
                for i in range(1, nb + 1):
                    if self.diffs[i] == diff:
                        d = i
                        break
                if d == 0:
                    d = nb + 1
                    self.diffs[d] = diff
                self.fonts[fontkey]["diff"] = d
            filename = font_dict.get("filename")
            if filename:
                if font_dict["type"] == "TrueType":
                    originalsize = font_dict["originalsize"]
                    self.font_files[filename] = {"length1": originalsize}
                else:
                    self.font_files[filename] = {
                        "length1": font_dict["size1"],
                        "length2": font_dict["size2"],
                    }

    def set_font(self, family, style="", size=0):
        "Select a font; size given in points"
        family = family.lower()
        if family == "":
            family = self.font_family
        if family == "arial":
            family = "helvetica"
        elif family == "symbol" or family == "zapfdingbats":
            style = ""
        style = style.upper()
        if "U" in style:
            self.underline = 1
            style = style.replace("U", "")
        else:
            self.underline = 0
        if style == "IB":
            style = "BI"
        if size == 0:
            size = self.font_size_pt
        # Test if font is already selected
        if (
            self.font_family == family
            and self.font_style == style
            and self.font_size_pt == size
        ):
            return
        # Test if used for the first time
        fontkey = family + style
        if fontkey not in self.fonts:
            # Check if one of the standard fonts
            if fontkey in self.core_fonts:
                if fontkey not in fpdf_charwidths:
                    # Load metric file
                    name = os.path.join(FPDF_FONT_DIR, family)
                    if family == "times" or family == "helvetica":
                        name += style.lower()
                    with open(name + ".font") as file:
                        exec(compile(file.read(), name + ".font", "exec"))
                    if fontkey not in fpdf_charwidths:
                        self.error("Could not include font metric file for" + fontkey)
                i = len(self.fonts) + 1
                self.fonts[fontkey] = {
                    "i": i,
                    "type": "core",
                    "name": self.core_fonts[fontkey],
                    "up": -100,
                    "ut": 50,
                    "cw": fpdf_charwidths[fontkey],
                }
            else:
                self.error("Undefined font: " + family + " " + style)
        # Select it
        self.font_family = family
        self.font_style = style
        self.font_size_pt = size
        self.font_size = size / self.k
        self.current_font = self.fonts[fontkey]
        self.unifontsubset = self.fonts[fontkey]["type"] == "TTF"
        if self.page > 0:
            self._out(
                sprintf("BT /F%d %.2f Tf ET", self.current_font["i"], self.font_size_pt)
            )

    def set_font_size(self, size):
        "Set font size in points"
        if self.font_size_pt == size:
            return
        self.font_size_pt = size
        self.font_size = size / self.k
        if self.page > 0:
            self._out(
                sprintf("BT /F%d %.2f Tf ET", self.current_font["i"], self.font_size_pt)
            )

    def set_stretching(self, factor):
        "Set from stretch factor percents (default: 100.0)"
        if self.font_stretching == factor:
            return
        self.font_stretching = factor
        if self.page > 0:
            self._out(sprintf("BT %.2f Tz ET", self.font_stretching))

    def add_link(self):
        "Create a new internal link"
        n = len(self.links) + 1
        self.links[n] = (0, 0)
        return n

    def set_link(self, link, y=0, page=-1):
        "Set destination of internal link"
        if y == -1:
            y = self.y
        if page == -1:
            page = self.page
        self.links[link] = [page, y]

    def link(self, x, y, w, h, link):
        "Put a link on the page"
        if not self.page in self.page_links:
            self.page_links[self.page] = []
        self.page_links[self.page] += [
            (x * self.k, self.h_pt - y * self.k, w * self.k, h * self.k, link),
        ]

    @check_page
    def text(self, x, y, txt=""):
        "Output a string"
        txt = self.normalize_text(txt)
        if self.unifontsubset:
            txt2 = self._escape(UTF8ToUTF16BE(txt, False))
            for uni in UTF8StringToArray(txt):
                self.current_font["subset"].append(uni)
        else:
            txt2 = self._escape(txt)
        s = sprintf(
            "BT %.2f %.2f Td (%s) Tj ET", x * self.k, (self.h - y) * self.k, txt2
        )
        if self.underline and txt != "":
            s += " " + self._dounderline(x, y, txt)
        if self.color_flag:
            s = "q " + self.text_color + " " + s + " Q"
        self._out(s)

    @check_page
    def rotate(self, angle, x=None, y=None):
        if x is None:
            x = self.x
        if y is None:
            y = self.y
        if self.angle != 0:
            self._out("Q")
        self.angle = angle
        if angle != 0:
            angle *= math.pi / 180
            c = math.cos(angle)
            s = math.sin(angle)
            cx = x * self.k
            cy = (self.h - y) * self.k
            s = sprintf(
                "q %.5F %.5F %.5F %.5F %.2F %.2F cm 1 0 0 1 %.2F %.2F cm",
                c,
                s,
                -s,
                c,
                cx,
                cy,
                -cx,
                -cy,
            )
            self._out(s)

    def accept_page_break(self):
        "Accept automatic page break or not"
        return self.auto_page_break

    @check_page
    def cell(self, w, h=0, txt="", border=0, ln=0, align="", fill=0, link=""):
        "Output a cell"
        txt = self.normalize_text(txt)
        k = self.k
        if (
            self.y + h > self.page_break_trigger
            and not self.in_footer
            and self.accept_page_break()
        ):
            # Automatic page break
            x = self.x
            ws = self.ws
            if ws > 0:
                self.ws = 0
                self._out("0 Tw")
            self.add_page(same=True)
            self.x = x
            if ws > 0:
                self.ws = ws
                self._out(sprintf("%.3f Tw", ws * k))
        if w == 0:
            w = self.w - self.r_margin - self.x
        s = ""
        if fill == 1 or border == 1:
            if fill == 1:
                if border == 1:
                    op = "B"
                else:
                    op = "f"
            else:
                op = "S"
            s = sprintf(
                "%.2f %.2f %.2f %.2f re %s ",
                self.x * k,
                (self.h - self.y) * k,
                w * k,
                -h * k,
                op,
            )
        if isinstance(border, basestring):
            x = self.x
            y = self.y
            if "L" in border:
                s += sprintf(
                    "%.2f %.2f m %.2f %.2f l S ",
                    x * k,
                    (self.h - y) * k,
                    x * k,
                    (self.h - (y + h)) * k,
                )
            if "T" in border:
                s += sprintf(
                    "%.2f %.2f m %.2f %.2f l S ",
                    x * k,
                    (self.h - y) * k,
                    (x + w) * k,
                    (self.h - y) * k,
                )
            if "R" in border:
                s += sprintf(
                    "%.2f %.2f m %.2f %.2f l S ",
                    (x + w) * k,
                    (self.h - y) * k,
                    (x + w) * k,
                    (self.h - (y + h)) * k,
                )
            if "B" in border:
                s += sprintf(
                    "%.2f %.2f m %.2f %.2f l S ",
                    x * k,
                    (self.h - (y + h)) * k,
                    (x + w) * k,
                    (self.h - (y + h)) * k,
                )
        if txt != "":
            if align == "R":
                dx = w - self.c_margin - self.get_string_width(txt, True)
            elif align == "C":
                dx = (w - self.get_string_width(txt, True)) / 2.0
            else:
                dx = self.c_margin
            if self.color_flag:
                s += "q " + self.text_color + " "

            # If multibyte, Tw has no effect - do word spacing using an adjustment before each space
            if self.ws and self.unifontsubset:
                for uni in UTF8StringToArray(txt):
                    self.current_font["subset"].append(uni)
                space = self._escape(UTF8ToUTF16BE(" ", False))
                s += sprintf(
                    "BT 0 Tw %.2F %.2F Td [",
                    (self.x + dx) * k,
                    (self.h - (self.y + 0.5 * h + 0.3 * self.font_size)) * k,
                )
                t = txt.split(" ")
                numt = len(t)
                for i in range(numt):
                    tx = t[i]
                    tx = "(" + self._escape(UTF8ToUTF16BE(tx, False)) + ")"
                    s += sprintf("%s ", tx)
                    if (i + 1) < numt:
                        adj = -(self.ws * self.k) * 1000 / self.font_size_pt
                        s += sprintf("%d(%s) ", adj, space)
                s += "] TJ"
                s += " ET"
            else:
                if self.unifontsubset:
                    txt2 = self._escape(UTF8ToUTF16BE(txt, False))
                    for uni in UTF8StringToArray(txt):
                        self.current_font["subset"].append(uni)
                else:
                    txt2 = self._escape(txt)
                s += sprintf(
                    "BT %.2f %.2f Td (%s) Tj ET",
                    (self.x + dx) * k,
                    (self.h - (self.y + 0.5 * h + 0.3 * self.font_size)) * k,
                    txt2,
                )

            if self.underline:
                s += " " + self._dounderline(
                    self.x + dx, self.y + 0.5 * h + 0.3 * self.font_size, txt
                )
            if self.color_flag:
                s += " Q"
            if link:
                self.link(
                    self.x + dx,
                    self.y + 0.5 * h - 0.5 * self.font_size,
                    self.get_string_width(txt, True),
                    self.font_size,
                    link,
                )
        if s:
            self._out(s)
        self.lasth = h
        if ln > 0:
            # Go to next line
            self.y += h
            if ln == 1:
                self.x = self.l_margin
        else:
            self.x += w

    @check_page
    def multi_cell(self, w, h, txt="", border=0, align="J", fill=0, split_only=False):
        "Output text with automatic or explicit line breaks"
        txt = self.normalize_text(txt)
        ret = []  # if split_only = True, returns splited text cells
        cw = self.current_font["cw"]
        if w == 0:
            w = self.w - self.r_margin - self.x
        wmax = (w - 2 * self.c_margin) * 1000.0 / self.font_size
        s = txt.replace("\r", "")
        nb = len(s)
        if nb > 0 and s[nb - 1] == "\n":
            nb -= 1
        b = 0
        if border:
            if border == 1:
                border = "LTRB"
                b = "LRT"
                b2 = "LR"
            else:
                b2 = ""
                if "L" in border:
                    b2 += "L"
                if "R" in border:
                    b2 += "R"
                if "T" in border:
                    b = b2 + "T"
                else:
                    b = b2
        sep = -1
        i = 0
        j = 0
        l = 0
        ns = 0
        nl = 1
        while i < nb:
            # Get next character
            c = s[i]
            if c == "\n":
                # Explicit line break
                if self.ws > 0:
                    self.ws = 0
                    if not split_only:
                        self._out("0 Tw")
                if not split_only:
                    self.cell(w, h, substr(s, j, i - j), b, 2, align, fill)
                else:
                    ret.append(substr(s, j, i - j))
                i += 1
                sep = -1
                j = i
                l = 0
                ns = 0
                nl += 1
                if border and nl == 2:
                    b = b2
                continue
            if c == " ":
                sep = i
                ls = l
                ns += 1
            if self.unifontsubset:
                l += self.get_string_width(c, True) / self.font_size * 1000.0
            else:
                l += cw.get(c, 0)
            if l > wmax:
                # Automatic line break
                if sep == -1:
                    if i == j:
                        i += 1
                    if self.ws > 0:
                        self.ws = 0
                        if not split_only:
                            self._out("0 Tw")
                    if not split_only:
                        self.cell(w, h, substr(s, j, i - j), b, 2, align, fill)
                    else:
                        ret.append(substr(s, j, i - j))
                else:
                    if align == "J":
                        if ns > 1:
                            self.ws = (wmax - ls) / 1000.0 * self.font_size / (ns - 1)
                        else:
                            self.ws = 0
                        if not split_only:
                            self._out(sprintf("%.3f Tw", self.ws * self.k))
                    if not split_only:
                        self.cell(w, h, substr(s, j, sep - j), b, 2, align, fill)
                    else:
                        ret.append(substr(s, j, sep - j))
                    i = sep + 1
                sep = -1
                j = i
                l = 0
                ns = 0
                nl += 1
                if border and nl == 2:
                    b = b2
            else:
                i += 1
        # Last chunk
        if self.ws > 0:
            self.ws = 0
            if not split_only:
                self._out("0 Tw")
        if border and "B" in border:
            b += "B"
        if not split_only:
            self.cell(w, h, substr(s, j, i - j), b, 2, align, fill)
            self.x = self.l_margin
        else:
            ret.append(substr(s, j, i - j))
        return ret

    @check_page
    def write(self, h, txt="", link=""):
        "Output text in flowing mode"
        txt = self.normalize_text(txt)
        cw = self.current_font["cw"]
        w = self.w - self.r_margin - self.x
        wmax = (w - 2 * self.c_margin) * 1000.0 / self.font_size
        s = txt.replace("\r", "")
        nb = len(s)
        sep = -1
        i = 0
        j = 0
        l = 0
        nl = 1
        while i < nb:
            # Get next character
            c = s[i]
            if c == "\n":
                # Explicit line break
                self.cell(w, h, substr(s, j, i - j), 0, 2, "", 0, link)
                i += 1
                sep = -1
                j = i
                l = 0
                if nl == 1:
                    self.x = self.l_margin
                    w = self.w - self.r_margin - self.x
                    wmax = (w - 2 * self.c_margin) * 1000.0 / self.font_size
                nl += 1
                continue
            if c == " ":
                sep = i
            if self.unifontsubset:
                l += self.get_string_width(c, True) / self.font_size * 1000.0
            else:
                l += cw.get(c, 0)
            if l > wmax:
                # Automatic line break
                if sep == -1:
                    if self.x > self.l_margin:
                        # Move to next line
                        self.x = self.l_margin
                        self.y += h
                        w = self.w - self.r_margin - self.x
                        wmax = (w - 2 * self.c_margin) * 1000.0 / self.font_size
                        i += 1
                        nl += 1
                        continue
                    if i == j:
                        i += 1
                    self.cell(w, h, substr(s, j, i - j), 0, 2, "", 0, link)
                else:
                    self.cell(w, h, substr(s, j, sep - j), 0, 2, "", 0, link)
                    i = sep + 1
                sep = -1
                j = i
                l = 0
                if nl == 1:
                    self.x = self.l_margin
                    w = self.w - self.r_margin - self.x
                    wmax = (w - 2 * self.c_margin) * 1000.0 / self.font_size
                nl += 1
            else:
                i += 1
        # Last chunk
        if i != j:
            self.cell(l / 1000.0 * self.font_size, h, substr(s, j), 0, 0, "", 0, link)

    @check_page
    def image(
        self,
        name,
        x=None,
        y=None,
        w=0,
        h=0,
        type="",
        link="",
        is_mask=False,
        mask_image=None,
    ):
        "Put an image on the page"
        if not name in self.images:
            # First use of image, get info
            if type == "":
                pos = name.rfind(".")
                if not pos:
                    self.error(
                        "image file has no extension and no type was specified: " + name
                    )
                type = substr(name, pos + 1)
            type = type.lower()
            if type == "jpg" or type == "jpeg":
                info = self._parsejpg(name)
            elif type == "png":
                info = self._parsepng(name)
            else:
                # Allow for additional formats
                # maybe the image is not showing the correct extension,
                # but the header is OK,
                succeed_parsing = False
                # try all the parsing functions
                parsing_functions = [self._parsejpg, self._parsepng, self._parsegif]
                for pf in parsing_functions:
                    try:
                        info = pf(name)
                        succeed_parsing = True
                        break
                    except:
                        pass
                # last resource
                if not succeed_parsing:
                    mtd = "_parse" + type
                    if not hasattr(self, mtd):
                        self.error("Unsupported image type: " + type)
                    info = getattr(self, mtd)(name)
                mtd = "_parse" + type
                if not hasattr(self, mtd):
                    self.error("Unsupported image type: " + type)
                info = getattr(self, mtd)(name)
            info["i"] = len(self.images) + 1
            # is_mask and mask_image
            if is_mask and info["cs"] != "DeviceGray":
                self.error("Mask must be a gray scale image")
            if mask_image:
                info["masked"] = mask_image
            self.images[name] = info
        else:
            info = self.images[name]
        # Automatic width and height calculation if needed
        if w == 0 and h == 0:
            # Put image at 72 dpi
            w = info["w"] / self.k
            h = info["h"] / self.k
        elif w == 0:
            w = h * info["w"] / info["h"]
        elif h == 0:
            h = w * info["h"] / info["w"]
        # Flowing mode
        if y is None:
            if (
                self.y + h > self.page_break_trigger
                and not self.in_footer
                and self.accept_page_break()
            ):
                # Automatic page break
                x = self.x
                self.add_page(same=True)
                self.x = x
            y = self.y
            self.y += h
        if x is None:
            x = self.x
        if not is_mask:
            self._out(
                sprintf(
                    "q %.2f 0 0 %.2f %.2f %.2f cm /I%d Do Q",
                    w * self.k,
                    h * self.k,
                    x * self.k,
                    (self.h - (y + h)) * self.k,
                    info["i"],
                )
            )
        if link:
            self.link(x, y, w, h, link)

        return info

    @check_page
    def ln(self, h=""):
        "Line Feed; default value is last cell height"
        self.x = self.l_margin
        if isinstance(h, basestring):
            self.y += self.lasth
        else:
            self.y += h

    def get_x(self):
        "Get x position"
        return self.x

    def set_x(self, x):
        "Set x position"
        if x >= 0:
            self.x = x
        else:
            self.x = self.w + x

    def get_y(self):
        "Get y position"
        return self.y

    def set_y(self, y):
        "Set y position and reset x"
        self.x = self.l_margin
        if y >= 0:
            self.y = y
        else:
            self.y = self.h + y

    def set_xy(self, x, y):
        "Set x and y positions"
        self.set_y(y)
        self.set_x(x)

    def output(self, name="", dest=""):
        """Output PDF to some destination

        By default the PDF is written to sys.stdout. If a name is given, the
        PDF is written to a new file. If dest='S' is given, the PDF data is
        returned as a byte string."""

        # Finish document if necessary
        if self.state < 3:
            self.close()
        dest = dest.upper()
        if dest == "":
            if name == "":
                dest = "I"
            else:
                dest = "F"
        if PY3K:
            # manage binary data as latin1 until PEP461 or similar is implemented
            buffer = self.buffer.encode("latin1")
        else:
            buffer = self.buffer
        if dest in ("I", "D"):
            # Python < 3 writes byte data transparently without "buffer"
            stdout = getattr(sys.stdout, "buffer", sys.stdout)
            stdout.write(buffer)
        elif dest == "F":
            # Save to local file
            with open(name, "wb") as f:
                f.write(buffer)
        elif dest == "S":
            # Return as a byte string
            return buffer
        else:
            self.error("Incorrect output destination: " + dest)

    def normalize_text(self, txt):
        "Check that text input is in the correct format/encoding"
        # - for TTF unicode fonts: unicode object (utf8 encoding)
        # - for built-in fonts: string instances (encoding: latin-1, cp1252)
        if not PY3K:
            if self.unifontsubset and isinstance(txt, str):
                return txt.decode("utf-8")
            elif not self.unifontsubset and isinstance(txt, unicode):
                return txt.encode(self.core_fonts_encoding)
        else:
            if not self.unifontsubset and self.core_fonts_encoding:
                return txt.encode(self.core_fonts_encoding).decode("latin-1")
        return txt

    def _dochecks(self):
        # Check for locale-related bug
        #        if(1.1==1):
        #            self.error("Don\'t alter the locale before including class file");
        # Check for decimal separator
        if sprintf("%.1f", 1.0) != "1.0":
            import locale

            locale.setlocale(locale.LC_NUMERIC, "C")

    def _getfontpath(self):
        return FPDF_FONT_DIR + "/"

    def _putpages(self):
        nb = self.page
        if hasattr(self, "str_alias_nb_pages"):
            # Replace number of pages in fonts using subsets (unicode)
            alias = UTF8ToUTF16BE(self.str_alias_nb_pages, False)
            r = UTF8ToUTF16BE(str(nb), False)
            for n in range(1, nb + 1):
                self.pages[n]["content"] = self.pages[n]["content"].replace(alias, r)
            # Now repeat for no pages in non-subset fonts
            for n in range(1, nb + 1):
                self.pages[n]["content"] = self.pages[n]["content"].replace(
                    self.str_alias_nb_pages, str(nb)
                )
        if self.def_orientation == "P":
            dw_pt = self.dw_pt
            dh_pt = self.dh_pt
        else:
            dw_pt = self.dh_pt
            dh_pt = self.dw_pt
        if self.compress:
            filter = "/Filter /FlateDecode "
        else:
            filter = ""
        for n in range(1, nb + 1):
            # Page
            self._newobj()
            self._out("<</Type /Page")
            self._out("/Parent 1 0 R")
            w_pt = self.pages[n]["w_pt"]
            h_pt = self.pages[n]["h_pt"]
            if w_pt != dw_pt or h_pt != dh_pt:
                self._out(sprintf("/MediaBox [0 0 %.2f %.2f]", w_pt, h_pt))
            self._out("/Resources 2 0 R")
            if self.page_links and n in self.page_links:
                # Links
                annots = "/Annots ["
                for pl in self.page_links[n]:
                    rect = sprintf(
                        "%.2f %.2f %.2f %.2f",
                        pl[0],
                        pl[1],
                        pl[0] + pl[2],
                        pl[1] - pl[3],
                    )
                    annots += (
                        "<</Type /Annot /Subtype /Link /Rect ["
                        + rect
                        + "] /Border [0 0 0] "
                    )
                    if isinstance(pl[4], basestring):
                        annots += (
                            "/A <</S /URI /URI " + self._textstring(pl[4]) + ">>>>"
                        )
                    else:
                        l = self.links[pl[4]]
                        if l[0] in self.orientation_changes:
                            h = w_pt
                        else:
                            h = h_pt
                        annots += sprintf(
                            "/Dest [%d 0 R /XYZ 0 %.2f null]>>",
                            1 + 2 * l[0],
                            h - l[1] * self.k,
                        )
                self._out(annots + "]")
            if self.pdf_version > "1.3":
                self._out("/Group <</Type /Group /S /Transparency" "/CS /DeviceRGB>>")
            self._out("/Contents " + str(self.n + 1) + " 0 R>>")
            self._out("endobj")
            # Page content
            content = self.pages[n]["content"]
            if self.compress:
                # manage binary data as latin1 until PEP461 or similar is implemented
                p = content.encode("latin1") if PY3K else content
                p = zlib.compress(p)
            else:
                p = content
            self._newobj()
            self._out("<<" + filter + "/Length " + str(len(p)) + ">>")
            self._putstream(p)
            self._out("endobj")
        # Pages root
        self.offsets[1] = len(self.buffer)
        self._out("1 0 obj")
        self._out("<</Type /Pages")
        kids = "/Kids ["
        for i in range(0, nb):
            kids += str(3 + 2 * i) + " 0 R "
        self._out(kids + "]")
        self._out("/Count " + str(nb))
        self._out(sprintf("/MediaBox [0 0 %.2f %.2f]", dw_pt, dh_pt))
        self._out(">>")
        self._out("endobj")

    def _putfonts(self):
        nf = self.n
        for diff in self.diffs:
            # Encodings
            self._newobj()
            self._out(
                "<</Type /Encoding /BaseEncoding /WinAnsiEncoding /Differences ["
                + self.diffs[diff]
                + "]>>"
            )
            self._out("endobj")
        for name, info in self.font_files.items():
            if "type" in info and info["type"] != "TTF":
                # Font file embedding
                self._newobj()
                self.font_files[name]["n"] = self.n
                with open(self._getfontpath() + name, "rb", 1) as f:
                    font = f.read()
                compressed = substr(name, -2) == ".z"
                if not compressed and "length2" in info:
                    header = ord(font[0]) == 128
                    if header:
                        # Strip first binary header
                        font = substr(font, 6)
                    if header and ord(font[info["length1"]]) == 128:
                        # Strip second binary header
                        font = substr(font, 0, info["length1"]) + substr(
                            font, info["length1"] + 6
                        )
                self._out("<</Length " + str(len(font)))
                if compressed:
                    self._out("/Filter /FlateDecode")
                self._out("/Length1 " + str(info["length1"]))
                if "length2" in info:
                    self._out("/Length2 " + str(info["length2"]) + " /Length3 0")
                self._out(">>")
                self._putstream(font)
                self._out("endobj")
        flist = [(x[1]["i"], x[0], x[1]) for x in self.fonts.items()]
        flist.sort()
        for idx, k, font in flist:
            # Font objects
            self.fonts[k]["n"] = self.n + 1
            type = font["type"]
            name = font["name"]
            if type == "core":
                # Standard font
                self._newobj()
                self._out("<</Type /Font")
                self._out("/BaseFont /" + name)
                self._out("/Subtype /Type1")
                if name != "Symbol" and name != "ZapfDingbats":
                    self._out("/Encoding /WinAnsiEncoding")
                self._out(">>")
                self._out("endobj")
            elif type == "Type1" or type == "TrueType":
                # Additional Type1 or TrueType font
                self._newobj()
                self._out("<</Type /Font")
                self._out("/BaseFont /" + name)
                self._out("/Subtype /" + type)
                self._out("/FirstChar 32 /LastChar 255")
                self._out("/Widths " + str(self.n + 1) + " 0 R")
                self._out("/FontDescriptor " + str(self.n + 2) + " 0 R")
                if font["enc"]:
                    if "diff" in font:
                        self._out("/Encoding " + str(nf + font["diff"]) + " 0 R")
                    else:
                        self._out("/Encoding /WinAnsiEncoding")
                self._out(">>")
                self._out("endobj")
                # Widths
                self._newobj()
                cw = font["cw"]
                s = "["
                for i in range(32, 256):
                    # Get doesn't raise exception; returns 0 instead of None if not set
                    s += str(cw.get(chr(i)) or 0) + " "
                self._out(s + "]")
                self._out("endobj")
                # Descriptor
                self._newobj()
                s = "<</Type /FontDescriptor /FontName /" + name
                for k in (
                    "Ascent",
                    "Descent",
                    "CapHeight",
                    "Flags",
                    "FontBBox",
                    "ItalicAngle",
                    "StemV",
                    "MissingWidth",
                ):
                    s += " /%s %s" % (k, font["desc"][k])
                filename = font["file"]
                if filename:
                    s += " /FontFile"
                    if type != "Type1":
                        s += "2"
                    s += " " + str(self.font_files[filename]["n"]) + " 0 R"
                self._out(s + ">>")
                self._out("endobj")
            elif type == "TTF":
                self.fonts[k]["n"] = self.n + 1
                ttf = TTFontFile()
                fontname = "MPDFAA" + "+" + font["name"]
                subset = font["subset"]
                del subset[0]
                ttfontstream = ttf.makeSubset(font["ttffile"], subset)
                ttfontsize = len(ttfontstream)
                fontstream = zlib.compress(ttfontstream)
                codeToGlyph = ttf.codeToGlyph
                ##del codeToGlyph[0]
                # Type0 Font
                # A composite font - a font composed of other fonts, organized hierarchically
                self._newobj()
                self._out("<</Type /Font")
                self._out("/Subtype /Type0")
                self._out("/BaseFont /" + fontname + "")
                self._out("/Encoding /Identity-H")
                self._out("/DescendantFonts [" + str(self.n + 1) + " 0 R]")
                self._out("/ToUnicode " + str(self.n + 2) + " 0 R")
                self._out(">>")
                self._out("endobj")

                # CIDFontType2
                # A CIDFont whose glyph descriptions are based on TrueType font technology
                self._newobj()
                self._out("<</Type /Font")
                self._out("/Subtype /CIDFontType2")
                self._out("/BaseFont /" + fontname + "")
                self._out("/CIDSystemInfo " + str(self.n + 2) + " 0 R")
                self._out("/FontDescriptor " + str(self.n + 3) + " 0 R")
                if font["desc"].get("MissingWidth"):
                    self._out("/DW %d" % font["desc"]["MissingWidth"])
                self._putTTfontwidths(font, ttf.maxUni)
                self._out("/CIDToGIDMap " + str(self.n + 4) + " 0 R")
                self._out(">>")
                self._out("endobj")

                # ToUnicode
                self._newobj()
                toUni = (
                    "/CIDInit /ProcSet findresource begin\n"
                    "12 dict begin\n"
                    "begincmap\n"
                    "/CIDSystemInfo\n"
                    "<</Registry (Adobe)\n"
                    "/Ordering (UCS)\n"
                    "/Supplement 0\n"
                    ">> def\n"
                    "/CMapName /Adobe-Identity-UCS def\n"
                    "/CMapType 2 def\n"
                    "1 begincodespacerange\n"
                    "<0000> <FFFF>\n"
                    "endcodespacerange\n"
                    "1 beginbfrange\n"
                    "<0000> <FFFF> <0000>\n"
                    "endbfrange\n"
                    "endcmap\n"
                    "CMapName currentdict /CMap defineresource pop\n"
                    "end\n"
                    "end"
                )
                self._out("<</Length " + str(len(toUni)) + ">>")
                self._putstream(toUni)
                self._out("endobj")

                # CIDSystemInfo dictionary
                self._newobj()
                self._out("<</Registry (Adobe)")
                self._out("/Ordering (UCS)")
                self._out("/Supplement 0")
                self._out(">>")
                self._out("endobj")

                # Font descriptor
                self._newobj()
                self._out("<</Type /FontDescriptor")
                self._out("/FontName /" + fontname)
                for kd in (
                    "Ascent",
                    "Descent",
                    "CapHeight",
                    "Flags",
                    "FontBBox",
                    "ItalicAngle",
                    "StemV",
                    "MissingWidth",
                ):
                    v = font["desc"][kd]
                    if kd == "Flags":
                        v = v | 4
                        v = v & ~32
                        # SYMBOLIC font flag
                    self._out(" /%s %s" % (kd, v))
                self._out("/FontFile2 " + str(self.n + 2) + " 0 R")
                self._out(">>")
                self._out("endobj")

                # Embed CIDToGIDMap
                # A specification of the mapping from CIDs to glyph indices
                cidtogidmap = ""
                cidtogidmap = ["\x00"] * 256 * 256 * 2
                for cc, glyph in codeToGlyph.items():
                    cidtogidmap[cc * 2] = chr(glyph >> 8)
                    cidtogidmap[cc * 2 + 1] = chr(glyph & 0xFF)
                cidtogidmap = "".join(cidtogidmap)
                if PY3K:
                    # manage binary data as latin1 until PEP461-like function is implemented
                    cidtogidmap = cidtogidmap.encode("latin1")
                cidtogidmap = zlib.compress(cidtogidmap)
                self._newobj()
                self._out("<</Length " + str(len(cidtogidmap)) + "")
                self._out("/Filter /FlateDecode")
                self._out(">>")
                self._putstream(cidtogidmap)
                self._out("endobj")

                # Font file
                self._newobj()
                self._out("<</Length " + str(len(fontstream)))
                self._out("/Filter /FlateDecode")
                self._out("/Length1 " + str(ttfontsize))
                self._out(">>")
                self._putstream(fontstream)
                self._out("endobj")
                del ttf
            else:
                # Allow for additional types
                mtd = "_put" + type.lower()
                if not method_exists(self, mtd):
                    self.error("Unsupported font type: " + type)
                self.mtd(font)

    def _putTTfontwidths(self, font, maxUni):
        if font["unifilename"]:
            cw127fname = os.path.splitext(font["unifilename"])[0] + ".cw127.pkl"
        else:
            cw127fname = None
        font_dict = load_cache(cw127fname)
        if font_dict is None:
            rangeid = 0
            range_ = {}
            range_interval = {}
            prevcid = -2
            prevwidth = -1
            interval = False
            startcid = 1
        else:
            rangeid = font_dict["rangeid"]
            range_ = font_dict["range"]
            prevcid = font_dict["prevcid"]
            prevwidth = font_dict["prevwidth"]
            interval = font_dict["interval"]
            range_interval = font_dict["range_interval"]
            startcid = 128
        cwlen = maxUni + 1

        # for each character
        subset = set(font["subset"])
        for cid in range(startcid, cwlen):
            if cid == 128 and cw127fname and not os.path.exists(cw127fname):
                try:
                    with open(cw127fname, "wb") as fh:
                        font_dict = {}
                        font_dict["rangeid"] = rangeid
                        font_dict["prevcid"] = prevcid
                        font_dict["prevwidth"] = prevwidth
                        font_dict["interval"] = interval
                        font_dict["range_interval"] = range_interval
                        font_dict["range"] = range_
                        pickle.dump(font_dict, fh)
                except IOError:
                    if not exception().errno == errno.EACCES:
                        raise  # Not a permission error.
            if cid > 255 and (cid not in subset):  #
                continue

            # start patch Kalkoken
            try:
                width = font["cw"][cid]
            except IndexError:
                width = 65535
            # end patch Kalkoken

            if width == 0:
                continue
            if width == 65535:
                width = 0
            if "dw" not in font or (font["dw"] and width != font["dw"]):
                if cid == (prevcid + 1):
                    if width == prevwidth:
                        if width == range_[rangeid][0]:
                            range_.setdefault(rangeid, []).append(width)
                        else:
                            range_[rangeid].pop()
                            # new range
                            rangeid = prevcid
                            range_[rangeid] = [prevwidth, width]
                        interval = True
                        range_interval[rangeid] = True
                    else:
                        if interval:
                            # new range
                            rangeid = cid
                            range_[rangeid] = [width]
                        else:
                            range_[rangeid].append(width)
                        interval = False
                else:
                    rangeid = cid
                    range_[rangeid] = [width]
                    interval = False
                prevcid = cid
                prevwidth = width
        prevk = -1
        nextk = -1
        prevint = False
        for k, ws in sorted(range_.items()):
            cws = len(ws)
            if k == nextk and not prevint and (not k in range_interval or cws < 3):
                if k in range_interval:
                    del range_interval[k]
                range_[prevk] = range_[prevk] + range_[k]
                del range_[k]
            else:
                prevk = k
            nextk = k + cws
            if k in range_interval:
                prevint = cws > 3
                del range_interval[k]
                nextk -= 1
            else:
                prevint = False
        w = []
        for k, ws in sorted(range_.items()):
            if len(set(ws)) == 1:
                w.append(" %s %s %s" % (k, k + len(ws) - 1, ws[0]))
            else:
                w.append(" %s [ %s ]\n" % (k, " ".join([str(int(h)) for h in ws])))  ##
        self._out("/W [%s]" % "".join(w))

    def _putimages(self):
        filter = ""
        if self.compress:
            filter = "/Filter /FlateDecode "
        i = [(x[1]["i"], x[1]) for x in self.images.items()]
        i.sort()
        for idx, info in i:
            self._putimage(info)
            del info["data"]
            if "smask" in info:
                del info["smask"]

    def _putimage(self, info):
        if "data" in info:
            self._newobj()
            info["n"] = self.n
            self._out("<</Type /XObject")
            self._out("/Subtype /Image")
            self._out("/Width " + str(info["w"]))
            self._out("/Height " + str(info["h"]))
            # set mask object for this image
            if "masked" in info:
                self._out("/SMask " + str(info["masked"]["n"] + 1) + " 0 R")

            if info["cs"] == "Indexed":
                self._out(
                    "/ColorSpace [/Indexed /DeviceRGB "
                    + str(len(info["pal"]) // 3 - 1)
                    + " "
                    + str(self.n + 1)
                    + " 0 R]"
                )
            else:
                self._out("/ColorSpace /" + info["cs"])
                if info["cs"] == "DeviceCMYK":
                    self._out("/Decode [1 0 1 0 1 0 1 0]")
            self._out("/BitsPerComponent " + str(info["bpc"]))
            if "f" in info:
                self._out("/Filter /" + info["f"])
            if "dp" in info:
                self._out("/DecodeParms <<" + info["dp"] + ">>")
            if "trns" in info and isinstance(info["trns"], list):
                trns = ""
                for i in range(0, len(info["trns"])):
                    trns += str(info["trns"][i]) + " " + str(info["trns"][i]) + " "
                self._out("/Mask [" + trns + "]")
            if "smask" in info:
                self._out("/SMask " + str(self.n + 1) + " 0 R")
            self._out("/Length " + str(len(info["data"])) + ">>")
            self._putstream(info["data"])
            self._out("endobj")
            # Soft mask
            if "smask" in info:
                dp = "/Predictor 15 /Colors 1 /BitsPerComponent 8 /Columns " + str(
                    info["w"]
                )
                smask = {
                    "w": info["w"],
                    "h": info["h"],
                    "cs": "DeviceGray",
                    "bpc": 8,
                    "f": info["f"],
                    "dp": dp,
                    "data": info["smask"],
                }
                self._putimage(smask)
            # Palette
            if info["cs"] == "Indexed":
                self._newobj()
                filter = self.compress and "/Filter /FlateDecode " or ""
                if self.compress:
                    pal = zlib.compress(info["pal"])
                else:
                    pal = info["pal"]
                self._out("<<" + filter + "/Length " + str(len(pal)) + ">>")
                self._putstream(pal)
                self._out("endobj")

    def _putxobjectdict(self):
        i = [(x["i"], x["n"]) for x in self.images.values()]
        i.sort()
        for idx, n in i:
            self._out("/I" + str(idx) + " " + str(n) + " 0 R")

    def _putresourcedict(self):
        self._out("/ProcSet [/PDF /Text /ImageB /ImageC /ImageI]")
        self._out("/Font <<")
        f = [(x["i"], x["n"]) for x in self.fonts.values()]
        f.sort()
        for idx, n in f:
            self._out("/F" + str(idx) + " " + str(n) + " 0 R")
        self._out(">>")
        self._out("/XObject <<")
        self._putxobjectdict()
        self._out(">>")

    def _putresources(self):
        self._putfonts()
        self._putimages()
        # Resource dictionary
        self.offsets[2] = len(self.buffer)
        self._out("2 0 obj")
        self._out("<<")
        self._putresourcedict()
        self._out(">>")
        self._out("endobj")

    def _putinfo(self):
        self._out(
            "/Producer "
            + self._textstring(
                "PyFPDF " + FPDF_VERSION + " http://pyfpdf.googlecode.com/"
            )
        )
        if hasattr(self, "title"):
            self._out("/Title " + self._textstring(self.title))
        if hasattr(self, "subject"):
            self._out("/Subject " + self._textstring(self.subject))
        if hasattr(self, "author"):
            self._out("/Author " + self._textstring(self.author))
        if hasattr(self, "keywords"):
            self._out("/Keywords " + self._textstring(self.keywords))
        if hasattr(self, "creator"):
            self._out("/Creator " + self._textstring(self.creator))
        self._out(
            "/CreationDate "
            + self._textstring("D:" + datetime.now().strftime("%Y%m%d%H%M%S"))
        )

    def _putcatalog(self):
        self._out("/Type /Catalog")
        self._out("/Pages 1 0 R")
        if self.zoom_mode == "fullpage":
            self._out("/OpenAction [3 0 R /Fit]")
        elif self.zoom_mode == "fullwidth":
            self._out("/OpenAction [3 0 R /FitH null]")
        elif self.zoom_mode == "real":
            self._out("/OpenAction [3 0 R /XYZ null null 1]")
        elif not isinstance(self.zoom_mode, basestring):
            self._out(
                sprintf("/OpenAction [3 0 R /XYZ null null %s]", self.zoom_mode / 100)
            )
        if self.layout_mode == "single":
            self._out("/PageLayout /SinglePage")
        elif self.layout_mode == "continuous":
            self._out("/PageLayout /OneColumn")
        elif self.layout_mode == "two":
            self._out("/PageLayout /TwoColumnLeft")

    def _putheader(self):
        self._out("%PDF-" + self.pdf_version)

    def _puttrailer(self):
        self._out("/Size " + str(self.n + 1))
        self._out("/Root " + str(self.n) + " 0 R")
        self._out("/Info " + str(self.n - 1) + " 0 R")

    def _enddoc(self):
        self._putheader()
        self._putpages()
        self._putresources()
        # Info
        self._newobj()
        self._out("<<")
        self._putinfo()
        self._out(">>")
        self._out("endobj")
        # Catalog
        self._newobj()
        self._out("<<")
        self._putcatalog()
        self._out(">>")
        self._out("endobj")
        # Cross-ref
        o = len(self.buffer)
        self._out("xref")
        self._out("0 " + (str(self.n + 1)))
        self._out("0000000000 65535 f ")
        for i in range(1, self.n + 1):
            self._out(sprintf("%010d 00000 n ", self.offsets[i]))
        # Trailer
        self._out("trailer")
        self._out("<<")
        self._puttrailer()
        self._out(">>")
        self._out("startxref")
        self._out(o)
        self._out("%%EOF")
        self.state = 3

    def _beginpage(self, orientation, format, same):
        self.page += 1
        self.pages[self.page] = {"content": ""}
        self.state = 2
        self.x = self.l_margin
        self.y = self.t_margin
        self.font_family = ""
        self.font_stretching = 100
        if not same:
            # Page format
            if format:
                # Change page format
                self.fw_pt, self.fh_pt = self.get_page_format(format, self.k)
            else:
                # Set to default format
                self.fw_pt = self.dw_pt
                self.fh_pt = self.dh_pt
            self.fw = self.fw_pt / self.k
            self.fh = self.fh_pt / self.k
            # Page orientation
            if not orientation:
                orientation = self.def_orientation
            else:
                orientation = orientation[0].upper()
            if orientation == "P":
                self.w_pt = self.fw_pt
                self.h_pt = self.fh_pt
            else:
                self.w_pt = self.fh_pt
                self.h_pt = self.fw_pt
            self.w = self.w_pt / self.k
            self.h = self.h_pt / self.k
            self.cur_orientation = orientation
            self.page_break_trigger = self.h - self.b_margin
            self.cur_orientation = orientation
        self.pages[self.page]["w_pt"] = self.w_pt
        self.pages[self.page]["h_pt"] = self.h_pt

    def _endpage(self):
        # End of page contents
        self.state = 1

    def _newobj(self):
        # Begin a new object
        self.n += 1
        self.offsets[self.n] = len(self.buffer)
        self._out(str(self.n) + " 0 obj")

    def _dounderline(self, x, y, txt):
        # Underline text
        up = self.current_font["up"]
        ut = self.current_font["ut"]
        w = self.get_string_width(txt, True) + self.ws * txt.count(" ")
        return sprintf(
            "%.2f %.2f %.2f %.2f re f",
            x * self.k,
            (self.h - (y - up / 1000.0 * self.font_size)) * self.k,
            w * self.k,
            -ut / 1000.0 * self.font_size_pt,
        )

    def load_resource(self, reason, filename):
        "Load external file"
        # by default loading from network is allowed for all images
        if reason == "image":
            if filename.startswith("http://") or filename.startswith("https://"):
                f = BytesIO(urlopen(filename).read())
            else:
                f = open(filename, "rb")
            return f
        else:
            self.error('Unknown resource loading reason "%s"' % reason)

    def _parsejpg(self, filename):
        # Extract info from a JPEG file
        f = None
        try:
            f = self.load_resource("image", filename)
            while True:
                markerHigh, markerLow = struct.unpack("BB", f.read(2))
                if markerHigh != 0xFF or markerLow < 0xC0:
                    raise SyntaxError("No JPEG marker found")
                elif markerLow == 0xDA:  # SOS
                    raise SyntaxError("No JPEG SOF marker found")
                elif (
                    markerLow == 0xC8
                    or (markerLow >= 0xD0 and markerLow <= 0xD9)  # JPG
                    or (markerLow >= 0xF0 and markerLow <= 0xFD)  # RSTx
                ):  # JPGx
                    pass
                else:
                    (dataSize,) = struct.unpack(">H", f.read(2))
                    data = f.read(dataSize - 2) if dataSize > 2 else ""
                    if (
                        (markerLow >= 0xC0 and markerLow <= 0xC3)
                        or (markerLow >= 0xC5 and markerLow <= 0xC7)  # SOF0 - SOF3
                        or (markerLow >= 0xC9 and markerLow <= 0xCB)  # SOF4 - SOF7
                        or (markerLow >= 0xCD and markerLow <= 0xCF)  # SOF9 - SOF11
                    ):  # SOF13 - SOF15
                        bpc, height, width, layers = struct.unpack_from(">BHHB", data)
                        colspace = (
                            "DeviceRGB"
                            if layers == 3
                            else ("DeviceCMYK" if layers == 4 else "DeviceGray")
                        )
                        break
        except Exception:
            if f:
                f.close()
            self.error(
                "Missing or incorrect image file: %s. error: %s"
                % (filename, str(exception()))
            )

        with f:
            # Read whole file from the start
            f.seek(0)
            data = f.read()
        return {
            "w": width,
            "h": height,
            "cs": colspace,
            "bpc": bpc,
            "f": "DCTDecode",
            "data": data,
        }

    def _parsegif(self, filename):
        # Extract info from a GIF file (via PNG conversion)
        if Image is None:
            self.error("PIL is required for GIF support")
        try:
            im = Image.open(filename)
        except Exception:
            self.error(
                "Missing or incorrect image file: %s. error: %s"
                % (filename, str(exception()))
            )
        else:
            # Use temporary file
            with tempfile.NamedTemporaryFile(delete=False, suffix=".png") as f:
                tmp = f.name
            if "transparency" in im.info:
                im.save(tmp, transparency=im.info["transparency"])
            else:
                im.save(tmp)
            info = self._parsepng(tmp)
            os.unlink(tmp)
        return info

    def _parsepng(self, filename):
        # Extract info from a PNG file
        f = self.load_resource("image", filename)
        # Check signature
        magic = f.read(8).decode("latin1")
        signature = "\x89" + "PNG" + "\r" + "\n" + "\x1a" + "\n"
        if not PY3K:
            signature = signature.decode("latin1")
        if magic != signature:
            self.error("Not a PNG file: " + filename)
        # Read header chunk
        f.read(4)
        chunk = f.read(4).decode("latin1")
        if chunk != "IHDR":
            self.error("Incorrect PNG file: " + filename)
        w = self._freadint(f)
        h = self._freadint(f)
        bpc = ord(f.read(1))
        if bpc > 8:
            self.error("16-bit depth not supported: " + filename)
        ct = ord(f.read(1))
        if ct == 0 or ct == 4:
            colspace = "DeviceGray"
        elif ct == 2 or ct == 6:
            colspace = "DeviceRGB"
        elif ct == 3:
            colspace = "Indexed"
        else:
            self.error("Unknown color type: " + filename)
        if ord(f.read(1)) != 0:
            self.error("Unknown compression method: " + filename)
        if ord(f.read(1)) != 0:
            self.error("Unknown filter method: " + filename)
        if ord(f.read(1)) != 0:
            self.error("Interlacing not supported: " + filename)
        f.read(4)
        dp = "/Predictor 15 /Colors "
        if colspace == "DeviceRGB":
            dp += "3"
        else:
            dp += "1"
        dp += " /BitsPerComponent " + str(bpc) + " /Columns " + str(w) + ""
        # Scan chunks looking for palette, transparency and image data
        pal = ""
        trns = ""
        data = bytes() if PY3K else str()
        n = 1
        while n != None:
            n = self._freadint(f)
            type = f.read(4).decode("latin1")
            if type == "PLTE":
                # Read palette
                pal = f.read(n)
                f.read(4)
            elif type == "tRNS":
                # Read transparency info
                t = f.read(n)
                if ct == 0:
                    trns = [
                        ord(substr(t, 1, 1)),
                    ]
                elif ct == 2:
                    trns = [
                        ord(substr(t, 1, 1)),
                        ord(substr(t, 3, 1)),
                        ord(substr(t, 5, 1)),
                    ]
                else:
                    pos = t.find("\x00".encode("latin1"))
                    if pos != -1:
                        trns = [
                            pos,
                        ]
                f.read(4)
            elif type == "IDAT":
                # Read image data block
                data += f.read(n)
                f.read(4)
            elif type == "IEND":
                break
            else:
                f.read(n + 4)
        if colspace == "Indexed" and not pal:
            self.error("Missing palette in " + filename)
        f.close()
        info = {
            "w": w,
            "h": h,
            "cs": colspace,
            "bpc": bpc,
            "f": "FlateDecode",
            "dp": dp,
            "pal": pal,
            "trns": trns,
        }
        if ct >= 4:
            # Extract alpha channel
            data = zlib.decompress(data)
            color = b("")
            alpha = b("")
            if ct == 4:
                # Gray image
                length = 2 * w
                for i in range(h):
                    pos = (1 + length) * i
                    color += b(data[pos])
                    alpha += b(data[pos])
                    line = substr(data, pos + 1, length)
                    re_c = re.compile("(.).".encode("ascii"), flags=re.DOTALL)
                    re_a = re.compile(".(.)".encode("ascii"), flags=re.DOTALL)
                    color += re_c.sub(lambda m: m.group(1), line)
                    alpha += re_a.sub(lambda m: m.group(1), line)
            else:
                # RGB image
                length = 4 * w
                for i in range(h):
                    pos = (1 + length) * i
                    color += b(data[pos])
                    alpha += b(data[pos])
                    line = substr(data, pos + 1, length)
                    re_c = re.compile("(...).".encode("ascii"), flags=re.DOTALL)
                    re_a = re.compile("...(.)".encode("ascii"), flags=re.DOTALL)
                    color += re_c.sub(lambda m: m.group(1), line)
                    alpha += re_a.sub(lambda m: m.group(1), line)
            del data
            data = zlib.compress(color)
            info["smask"] = zlib.compress(alpha)
            if self.pdf_version < "1.4":
                self.pdf_version = "1.4"
        info["data"] = data
        return info

    def _freadint(self, f):
        # Read a 4-byte integer from file
        try:
            return struct.unpack(">I", f.read(4))[0]
        except:
            return None

    def _textstring(self, s):
        # Format a text string
        return "(" + self._escape(s) + ")"

    def _escape(self, s):
        # Add \ before \, ( and )
        return (
            s.replace("\\", "\\\\")
            .replace(")", "\\)")
            .replace("(", "\\(")
            .replace("\r", "\\r")
        )

    def _putstream(self, s):
        self._out("stream")
        self._out(s)
        self._out("endstream")

    def _out(self, s):
        # Add a line to the document
        if PY3K and isinstance(s, bytes):
            # manage binary data as latin1 until PEP461-like function is implemented
            s = s.decode("latin1")
        elif not PY3K and isinstance(s, unicode):
            s = s.encode("latin1")  # default encoding (font name and similar)
        elif not isinstance(s, basestring):
            s = str(s)
        if self.state == 2:
            self.pages[self.page]["content"] += s + "\n"
        else:
            self.buffer += s + "\n"

    @check_page
    def interleaved2of5(self, txt, x, y, w=1.0, h=10.0):
        "Barcode I2of5 (numeric), adds a 0 if odd lenght"
        narrow = w / 3.0
        wide = w

        # wide/narrow codes for the digits
        bar_char = {
            "0": "nnwwn",
            "1": "wnnnw",
            "2": "nwnnw",
            "3": "wwnnn",
            "4": "nnwnw",
            "5": "wnwnn",
            "6": "nwwnn",
            "7": "nnnww",
            "8": "wnnwn",
            "9": "nwnwn",
            "A": "nn",
            "Z": "wn",
        }

        self.set_fill_color(0)
        code = txt
        # add leading zero if code-length is odd
        if len(code) % 2 != 0:
            code = "0" + code

        # add start and stop codes
        code = "AA" + code.lower() + "ZA"

        for i in range(0, len(code), 2):
            # choose next pair of digits
            char_bar = code[i]
            char_space = code[i + 1]
            # check whether it is a valid digit
            if not char_bar in bar_char.keys():
                raise RuntimeError('Char "%s" invalid for I25: ' % char_bar)
            if not char_space in bar_char.keys():
                raise RuntimeError('Char "%s" invalid for I25: ' % char_space)

            # create a wide/narrow-seq (first digit=bars, second digit=spaces)
            seq = ""
            for s in range(0, len(bar_char[char_bar])):
                seq += bar_char[char_bar][s] + bar_char[char_space][s]

            for bar in range(0, len(seq)):
                # set line_width depending on value
                if seq[bar] == "n":
                    line_width = narrow
                else:
                    line_width = wide

                # draw every second value, the other is represented by space
                if bar % 2 == 0:
                    self.rect(x, y, line_width, h, "F")

                x += line_width

    @check_page
    def code39(self, txt, x, y, w=1.5, h=5.0):
        """Barcode 3of9"""
        dim = {"w": w, "n": w / 3.0}
        chars = {
            "0": "nnnwwnwnn",
            "1": "wnnwnnnnw",
            "2": "nnwwnnnnw",
            "3": "wnwwnnnnn",
            "4": "nnnwwnnnw",
            "5": "wnnwwnnnn",
            "6": "nnwwwnnnn",
            "7": "nnnwnnwnw",
            "8": "wnnwnnwnn",
            "9": "nnwwnnwnn",
            "A": "wnnnnwnnw",
            "B": "nnwnnwnnw",
            "C": "wnwnnwnnn",
            "D": "nnnnwwnnw",
            "E": "wnnnwwnnn",
            "F": "nnwnwwnnn",
            "G": "nnnnnwwnw",
            "H": "wnnnnwwnn",
            "I": "nnwnnwwnn",
            "J": "nnnnwwwnn",
            "K": "wnnnnnnww",
            "L": "nnwnnnnww",
            "M": "wnwnnnnwn",
            "N": "nnnnwnnww",
            "O": "wnnnwnnwn",
            "P": "nnwnwnnwn",
            "Q": "nnnnnnwww",
            "R": "wnnnnnwwn",
            "S": "nnwnnnwwn",
            "T": "nnnnwnwwn",
            "U": "wwnnnnnnw",
            "V": "nwwnnnnnw",
            "W": "wwwnnnnnn",
            "X": "nwnnwnnnw",
            "Y": "wwnnwnnnn",
            "Z": "nwwnwnnnn",
            "-": "nwnnnnwnw",
            ".": "wwnnnnwnn",
            " ": "nwwnnnwnn",
            "*": "nwnnwnwnn",
            "$": "nwnwnwnnn",
            "/": "nwnwnnnwn",
            "+": "nwnnnwnwn",
            "%": "nnnwnwnwn",
        }
        self.set_fill_color(0)
        for c in txt.upper():
            if c not in chars:
                raise RuntimeError('Invalid char "%s" for Code39' % c)
            for i, d in enumerate(chars[c]):
                if i % 2 == 0:
                    self.rect(x, y, dim[d], h, "F")
                x += dim[d]
            x += dim["n"]
